"""
Locator provides a key->Location store,
where a Location is a path+metadata specialization of Strang

"""

from ._interface import Location_p, Locator_p
from .location import Location
from .locator import JGDVLocator



"""
Some Enum's I Keep writing
"""
import enum

class LoopControl_e(enum.Enum):
    """
      Describes how to continue an accumulating loop.
      (like walking a a tree)

    yesAnd     : is a result, and try others.
    yes        : is a result, don't try others, Finish.
    noBut      : not a result, try others.
    no         : not a result, don't try others, Finish.
    """
    yesAnd  = enum.auto()  # noqa: N815
    yes     = enum.auto()
    noBut   = enum.auto()  # noqa: N815
    no      = enum.auto()

    @classmethod
    def loop_yes_set(cls) -> set:
        return  {cls.yesAnd, cls.yes, True}

    @classmethod
    def loop_no_set(cls) -> set:
        return  {cls.no, cls.noBut, False, None}

class CurrentState_e(enum.Enum):
    """
      Enumeration of the different states a task can be in.
    """
    TEARDOWN        = enum.auto()
    SUCCESS         = enum.auto()
    FAILED          = enum.auto()
    HALTED          = enum.auto()
    WAIT            = enum.auto()
    READY           = enum.auto()
    RUNNING         = enum.auto()
    EXISTS          = enum.auto()
    INIT            = enum.auto()

    DEFINED         = enum.auto()
    DECLARED        = enum.auto()
    ARTIFACT        = enum.auto()

class ActionResult_e(enum.Enum):
    """
      Enums for how a task can describe its response
    """
    SUCCEED  = enum.auto()
    FAIL     = enum.auto()
    SKIP     = enum.auto()
    HALT     = enum.auto()

class TaskPolicyEnum(enum.Flag):
    """
      Combinable Policy Types:
      breaker  : fails fast
      bulkhead : limits extent of problem and continues
      retry    : trys to do the action again to see if its resolved
      timeout  : waits then fails
      cache    : reuses old results
      fallback : uses defined alternatives
      cleanup  : uses defined cleanup actions
      debug    : triggers pdb
      pretend  : pretend everything went fine
      accept   : accept the failure

      breaker will overrule bulkhead
    """
    BREAKER  = enum.auto()
    BULKHEAD = enum.auto()
    RETRY    = enum.auto()
    TIMEOUT  = enum.auto()
    CACHE    = enum.auto()
    FALLBACK = enum.auto()
    CLEANUP  = enum.auto()
    DEBUG    = enum.auto()
    PRETEND  = enum.auto()
    ACCEPT   = enum.auto()



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import string
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv.structs.strang import CodeReference

# ##-- end 1st party imports

from .._interface import DKeyMark_e
from .._util._interface import ExpInst_d
from .. import DKey
from ..keys import MultiDKey, NonDKey, SingleDKey

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Ident, RxStr, Rx
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class PathDKey(DKey[list], mark=pl.Path):
    """
    A Simple key that always expands to a path, and is then normalised
    """
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data.convert         = "p"
        self.data.expansion_type  = pl.Path
        self.data.typecheck       = pl.Path


    def exp_final_h(self, inst:ExpInst_d, opts:dict) -> Maybe[ExpInst_d]:
        return ExpInst_d(value=pl.Path(inst.value).expanduser().resolve(),
                         literal=True
                         )



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref

# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Final,
    Generator,
    Generic,
    Iterable,
    Iterator,
    Mapping,
    Match,
    MutableMapping,
    Protocol,
    Sequence,
    Tuple,
    TypeAlias,
    TypeGuard,
    TypeVar,
    cast,
    final,
    overload,
    runtime_checkable,
)
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv import Maybe

# ##-- end 1st party imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# True to process, False to reject

class BlacklistFilter:
    """
      A Logging filter to blacklist regexs of logger names
    """

    def __init__(self, blacklist:list[str]=None):
        self._blacklist   = blacklist or []
        self.blacklist_re  = re.compile("^({})".format("|".join(self._blacklist)))

    def __call__(self, record) -> bool:
        if record.name == "root":
            return True
        if not bool(self._blacklist):
            return True

        return not bool(self.blacklist_re.match(record.name))



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple,
                    TypeVar, cast)
import warnings

##-- end imports

import pytest
from jgdv.structs.chainguard import ChainGuard

logging = logmod.root

data_dir = pl.Path(__file__).parent / "__data"

class TestGuardLoader:


    def test_initial(self):
        data_str = (data_dir / "data.toml").read_text()
        assert(bool(data_str))
        simple = ChainGuard.read(data_str)
        assert(bool(simple))

    def test_content(self):
        data_str = (data_dir / "data.toml").read_text()
        assert(bool(data_str))
        simple = ChainGuard.read(data_str)
        assert("basic" in simple)
        assert("value" in simple)

    def test_from_dict(self):
        simple = ChainGuard.from_dict({"val": 5, "other": "blah", "nested": {"val": True}})
        assert("val" in simple)
        assert("other" in simple)
        assert(simple.nested.val is True)

    def test_load(self):
        simple = ChainGuard.load(data_dir / "data.toml")
        assert("basic" in simple)
        assert(simple.basic == "test")

    def test_load_dir(self):
        simple = ChainGuard.load_dir(data_dir)
        assert("basic" in simple)
        assert("a-different-val" in simple)
        assert(simple.a_different_val == "blah")
        assert(simple.basic == "test")



 #!/usr/bin/env python3
"""

"""
# ruff: noqa: B019, PLR2004
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import functools as ftz
import importlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv import Mixin, Proto

# ##-- end 1st party imports

from .processor import StrangBasicProcessor
from .formatter import StrangFormatter
from . import errors
from . import _interface as API # noqa: N812
from ._meta import StrangMeta
from jgdv.mixins.annotate import SubAlias_m

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
import enum
from uuid import UUID
from collections.abc import Iterator

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
logging.disabled = False
##-- end logging

##--|

class _StrangSlicer:
    """
        Access sections and words of a Strang,
        by name or index.

        val = Strang('a.b.c::d.e.f')
        val[:]          -> str(a.b.c::d.e.f)
        val[0,:]        -> a.b.c
        val[0]          -> a.b.c
        val[0,0]        -> a
        val[0,:-1]      -> a.b
        val['head']     -> a.b.c
        val['head', -1] -> c
        val[:,:,:-1]    -> a.b.c::d.e
    """
    __slots__ = ()

    def getitem(self, obj:API.Strang_p, args:API.ItemIndex) -> str: # type: ignore[override]
        i       : int
        x       : Any
        words   : list[str]
        gotten  : str
        match self.discrim_getitem_args(obj, args):
            case Iterator() as sec_iter: # full expansion
                gotten = self.run_iterator(obj, sec_iter)
            case int()|slice() as section, None:
                bounds = obj.data.sections[section]
                gotten = API.STRGET(obj, bounds)
            case None, int() as flat:
                gotten = API.STRGET(obj, obj.data.words[flat])
            case None, slice() as flat:
                selection = obj.data.words[flat]
                gotten = API.STRGET(obj, slice(selection[0].start, selection[-1].stop, flat.step))
            case int() as section, int() as word:
                idx = obj.data.sec_words[section][word]
                gotten = API.STRGET(obj, obj.data.words[idx])
            case int() as section, slice() as word:
                case   = obj.section(section).case or ""
                words  = [API.STRGET(obj, obj.data.words[i]) for i in obj.data.sec_words[section][word]]
                gotten = case.join(words)
            case int()|slice() as basic:
                gotten = API.STRGET(obj, basic)
            case True, [*xs]:
                gotten = self.multi_slice(obj, xs)
            case _:
                raise KeyError(errors.UnkownSlice, args)
        ##--|
        return gotten

    def discrim_getitem_args(self, obj:API.Strang_p, args:API.ItemIndex) -> Iterator[API.Sec_d]|tuple[Maybe[API.ItemIndex], ...]|API.ItemIndex:
        result : Iterator|tuple|API.ItemIndex
        match args:
            case int() | slice() as x: # Normal str-like
                result = x
            case str() as k: # whole section by name
                result = obj.section(k).idx, None
            case [slice(), slice()] if not bool(obj.data.words):
                result = obj.data.sections[0]
            case [slice() as secs, slice(start=None, stop=None, step=None)]: # type: ignore[misc]
                sec_it = itz.islice(obj.sections(), secs.start, secs.stop, secs.step)
                result = sec_it
            case [int() as idx, *_] if len(obj.sections()) < idx:
                raise KeyError(errors.MissingSectionIndex.format(cls=obj.__class__.__name__,
                                                            idx=idx,
                                                            sections=len(obj.sections())))
            case [str() as key, *_] if key not in obj._sections.named:
                raise KeyError(errors.MissingSectionName.format(cls=obj.__class__.__name__,
                                                                key=key))
            case [slice() as secs, *subs] if secs.start is None and secs.stop is None:
                if len(subs) != len(obj.data.sec_words[secs]): # type: ignore[misc]
                    raise KeyError(errors.SliceMisMatch, len(subs), len(obj.data.sec_words[secs]))
                result = True, tuple(subs)
            case [str()|int() as i, slice()|int() as x]: # Section-word
                result = obj.section(i).idx, x
            case [None, slice()|int() as x]: # Flat slice
                result = None, x
            case x:
                raise TypeError(type(x), x)
        ##--|
        return result

    def run_iterator(self, obj:API.Strang_p, sec_iter:Iterator) -> str:
        sec : API.Sec_d
        result = []
        for sec in sec_iter:
            for word in obj.words(sec.idx, case=True):
                match word:
                    case UUID() as x:
                        result.append(f"<uuid:{x}>")
                    case x:
                        result.append(str(x))
            else:
                result.append(sec.end or "")
        else:
            return "".join(result)

    def multi_slice(self, obj:API.Strang_p, slices:Iterable) -> str:
        result = []
        for i,x in enumerate(slices):
            if x is None:
                continue
            result.append(obj[i,x])
            if (end:=obj.section(i).end) is not None:
                result.append(end)
        else:
            return "".join(result)
##--|

@Proto(API.Strang_p, mod_mro=False)
class Strang[*K](SubAlias_m, str, metaclass=StrangMeta, fresh_registry=True):
    """ A Structured String Baseclass.

    A Normal str, but is parsed on construction to extract and validate
    certain form and metadata.

    The Form of a Strang is::

        {group}{sep}{body}
        eg: group.val::body.val

    Body objs can be marks (Strang.bmark_e), and UUID's as well as str's

    strang[x] and strang[x:y] are changed to allow structured access::

        val         = Strang("a.b.c::d.e.f")
        val[0] # a.b.c
        val[1] # d.e.f

    """
    __slots__       = ("data", "meta")
    __match_args__  = ("head", "body")

    ##--|
    _processor  : ClassVar                 = StrangBasicProcessor()
    _formatter  : ClassVar                 = StrangFormatter()
    _slicer     : ClassVar[_StrangSlicer]  = _StrangSlicer()
    _sections   : ClassVar                 = API.STRANG_ALT_SECS

    data        : API.Strang_d
    meta        : dict

    @classmethod
    def sections(cls) -> API.Sections_d:
        return cls._sections

    @classmethod
    def section(cls, arg:int|str) -> API.Sec_d:
        return cls._sections[arg]

    ##--|

    def __init__(self, *args:Any, **kwargs:Any) -> None:  # noqa: ANN401, ARG002
        super().__init__()
        self.meta  = dict(kwargs)
        self.data  = API.Strang_d()

   ##--| dunders

    def __str__(self) -> str:
        """ Provides a fully expanded string

        eg: a.b.c::d.e.f..<uuid:{val}>
        """
        return format(self, "a+")

    def __repr__(self) -> str:
        body = self[:]
        cls  = self.__class__.__name__
        return f"<{cls}: {body}>"

    def __format__(self, spec:str) -> str:
        """ Basic formatting to get just a section """
        result : str
        match spec:
            case "a" | "a-" | "a+" if not self.data.args_start:
                result = self[:,:]
            case "a-": # No Args, No Expansion
                result = self[:self.data.args_start]
            case "a+" if self.data.args_start: # Full Args
                result = f"{self[:,:]}[<uuid:{self.uuid()}>]"
            case "a" if self.data.args_start: # Simple Args
                result = self[:]
            case "a=" if self.data.args_start: # only args
                result = self[self.data.args_start+1:-1]
            case "a=":
                result = ""
            case "u" if self.data.uuid:
                val = self.data.uuid
                result = f"<uuid:{val}>"
            case "u":
                raise ValueError()
            case _:
                result = super().__format__(spec)

        return result

    def __hash__(self) -> int:
        return str.__hash__(str(self))

    def __lt__(self:API.Strang_p, other:object) -> bool:
        match other:
            case API.Strang_p() | str() as x if not len(self) < len(x):
                logging.debug("Length mismatch")
                return False
            case API.Strang_p():
                pass
            case x:
                logging.debug("Type failure")
                return False

        assert(isinstance(self, API.Strang_p))
        assert(isinstance(other, API.Strang_p))
        if not self[0,:] == other[0,:]:
            logging.debug("head mismatch")
            return False

        for x,y in zip(self.words(1), other.words(1), strict=False):
            if x != y:
                logging.debug("Faileid on: %s : %s", x, y)
                return False

        return True

    def __le__(self:API.Strang_p, other:object) -> bool:
        match other:
            case API.Strang_p() as x:
                return hash(self) == hash(other) or (self < x) # type: ignore
            case str():
                return hash(self) == hash(other)
            case x:
                raise TypeError(type(x))

    def __eq__(self, other:object) -> bool:
        match other:
            case Strang() as x if self.uuid() and x.uuid():
                return hash(self) == hash(other)
            case UUID() as x:
                return self.uuid() == x
            case x if self.uuid():
                h_other = hash(x)
                return hash(self) == h_other or hash(self[:]) == h_other
            case x:
                return hash(self) == hash(x)

    def __ne__(self, other:object) -> bool:
        return not self == other

    def __iter__(self) -> Iterator:
        """ iterate over words """
        for sec in self.sections():
            yield from self.words(sec.idx)

    def __getitem__(self, args:API.ItemIndex) -> str: # type: ignore[override]
        """
        Access sections and words of a Strang,
        by name or index.

        val = Strang('a.b.c::d.e.f')
        val[:]          -> str(a.b.c::d.e.f)
        val[0,:]        -> a.b.c
        val[0]          -> a.b.c
        val[0,0]        -> a
        val[0,:-1]      -> a.b
        val['head']     -> a.b.c
        val['head', -1] -> c
        val[:,:,:-1]    -> a.b.c::d.e
        """
        return self._slicer.getitem(cast("API.Strang_p", self), args)

    def __getattr__(self, val:str) -> str:
        """ Enables using match statement for entire sections

        eg: case Strang(head=x, body=y):...

        """
        match val:
            case str() as x if x in self.sections():
                return self[val]
            case _:
                raise AttributeError(val)

    def __contains__(self:API.Strang_p, other:object) -> bool:
        """ test for conceptual containment of names
        other(a.b.c) ∈ self(a.b) ?
        ie: self < other
        """
        match other:
            case API.StrangMarkAbstract_e() as x:
                return x in self.data.meta
            case UUID() as x:
                return (x == self.uuid() or x in self.data.meta)
            case str() as needle:
                return API.STRCON(self, needle)
            case _:
                return False

    ##--| Properties

    @property
    def base(self) -> Self:
        return self

    @property
    def shape(self) -> tuple[int, ...]:
        return tuple(len(x) for x in self.data.sec_words)

    ##--| Access

    def index(self, *sub:API.FindSlice, start:Maybe[int]=None, end:Maybe[int]=None) -> int: # type: ignore[override]
        """ Extended str.index, to handle marks and word slices """
        needle  : str|API.StrangMarkAbstract_e
        word    : int
        match sub:
            case [API.StrangMarkAbstract_e() as mark]:
                idx = self.data.meta.index(mark)
                return self.data.words[idx].start
            case ["", *_]:
                raise ValueError(errors.IndexOfEmptyStr, sub)
            case [str() as needle]:
                pass
            case [str()|int() as sec, int() as word]:
                needle = self.get(sec, word)
            case _:
                raise TypeError(type(sub), sub)

        match needle:
            case API.StrangMarkAbstract_e():
                return self.index(needle, start=start, end=end)
            case _:
                return str.index(self, needle, start, end)

    def rindex(self, *sub:API.FindSlice, start:Maybe[int]=None, end:Maybe[int]=None) -> int: # type: ignore[override]
        """ Extended str.rindex, to handle marks and word slices """
        needle  : str
        word    : int
        match sub:
            case [API.StrangMarkAbstract_e() as mark]:
                word_idx  = max(-1, *(i for i,x in enumerate(self.data.meta) if x == mark))
                if word_idx == -1:
                    raise ValueError(mark)
                return self.data.words[word_idx].start
            case ["", *_]:
                raise ValueError(errors.IndexOfEmptyStr, sub)
            case [str() as needle]:
                pass
            case [int()|str() as sec, int() as word]:
                idx = self.section(sec).idx
                word_idx = self.data.sec_words[idx][word]
                return self.data.words[word_idx].start
            case x:
                raise ValueError(x)

        return str.rindex(self, needle, start, end)

    def get(self, *args:API.SectionIndex|API.WordIndex) -> Any:  # noqa: ANN401
        """ Accessor to get internal data """
        x     : Any
        sec   : int
        word  : int
        idx   : int
        match args:
            case [str() | int() as i]:
                return self[i]
            case [int() as sec, int() as word]:
                idx = self.data.sec_words[sec][word]
            case [str() as k, int() as word]:
                sec = self.section(k).idx
                idx = self.data.sec_words[sec][word]
            case x:
                raise KeyError(x)

        try:
            val = self.data.meta[idx] # type: ignore[index]
        except (ValueError, IndexError):
            return self[sec, word]
        else:
            match val:
                case None:
                    return self[sec,word]
                case _:
                    return val

    def words(self, idx:int|str, *, select:Maybe[slice]=None, case:bool=False) -> Iterator:
        """ Get the word values of a section.
        case=True adds the case in between values,
        select can be a slice that limits the returned values

        """
        count    : int
        gen      : Iterator
        section  : API.Sec_d
        sec_case : str
        section  = self.section(idx)
        sec_case = section.case or ""
        count    = len(self.data.sec_words[section.idx])
        if not bool(self.data.words):
            return
        if count == 0:
            return

        match select:
            case None:
                select = slice(None)
            case slice():
                pass

        gen       = itz.islice(range(count), select.start, select.stop, select.step)
        offbyone  = itz.tee(gen, 2)
        next(offbyone[1])

        for x,y in itz.zip_longest(*offbyone, fillvalue=None):
            yield self.get(section.idx, x)
            if case and y is not None:
                yield sec_case

    def args(self) -> Maybe[tuple]:
        return self.data.args
    ##--| Modify

    def push(self, *args:API.PushVal) -> Self:
        """ extend a strang with values

        Pushed onto the last section, with a section.marks.skip() mark first

        eg: val = Strang('a.b.c::d.e.f')
        val.push(val.section(1).mark.head) -> 'a.b.c::d.e.f..$head$'
        val.push(uuid=True) -> 'a.b.c::d.e.f..<uuid>'
        val.push(uuid=uuid1()) -> 'a.b.c::d.e.f..<uuid:{val}>'
        """
        word  : API.PushVal
        x     : API.PushVal
        words  = [format(self, "a-")]
        marks  = self.section(-1).marks or API.DefaultBodyMarks_e
        match marks.skip():
            case API.StrangMarkAbstract_e() as x:
                mark = x.value
                words.append(x.value)
            case _:
                raise ValueError(errors.NoSkipMark)

        for word in args:
            match word:
                case API.StrangMarkAbstract_e() as x if x in type(x).idempotent() and x in self:
                    pass
                case API.StrangMarkAbstract_e() as x if x in type(x).idempotent() and x in words:
                    pass
                case API.StrangMarkAbstract_e() as x if x in type(x).idempotent() and x in self:
                    words.append(x.value)
                case str() as x:
                    words.append(x)
                case UUID() as x:
                    words.append(f"<uuid:{x}>")
                case None:
                    words.append(mark)
                case x:
                    words.append(str(x))
        else:
            return self.__class__(*words)

    def pop(self, *, top:bool=True)-> Self:
        """
        Strip off one marker's worth of the name, or to the top marker.
        eg:
        root(test::a.b.c..<UUID>.sub..other) => test::a.b.c..<UUID>.sub
        root(test::a.b.c..<UUID>.sub..other, top=True) => test::a.b.c
        """
        mark = (self.section(-1).marks or API.DefaultBodyMarks_e).skip()
        assert(mark is not None)
        try:
            match top:
                case True:
                    next_mark = self.index(mark)
                case False:
                    next_mark = self.rindex(mark)
        except ValueError:
            return self
        else:
            return type(self)(self[:next_mark])

    def mark(self, mark:str|API.StrangMarkAbstract_e) -> Self:
        """ Add a given mark if it is last section appropriate  """
        appropriate = self.section(-1).marks
        assert(appropriate is not None)
        match mark:
            case str() as x if x in appropriate:
                return self.push(appropriate(x))
            case API.StrangMarkAbstract_e() as x if x in appropriate:
                return self.push(x)
            case x:
                raise ValueError(x)

    ##--| UUIDs

    def uuid(self) -> Maybe[UUID]:
        return self.data.uuid

    def to_uniq(self, *args:str) -> Self:
        """ Generate a concrete instance of this name with a UUID prepended,

          ie: a.task.group::task.name..{prefix?}.$gen$.<UUID>
        """
        match args:
            case [] if self.uuid():
                return self
            case [] if bool(curr:=f"{self:a=}"):
                return self.__class__(f"{self:a-}[{curr},<uuid>]")
            case []:
                return self.__class__(f"{self:a-}[<uuid>]")
            case [*xs, x] if bool(curr:=f"{self:a=}"):
                return self.__class__(f"{self:a-}", *xs, f"{x}[<uuid>]")
            case [x]:
                return self.__class__(f"{self:a-}", f"{x}[<uuid>]")
            case x:
                raise TypeError(type(x))

    def de_uniq(self) -> Self:
        """ a.b.c::d.e.f[<uuid>] -> a.b.c::d.e.f

        """
        match self.uuid():
            case None:
                return self
            case _:
                return self.__class__(f"{self[:,:]}")

    ##--| Other

    def format(self, *args:Any, **kwargs:Any) -> str:  # noqa: ANN401
        """ Advanced formatting for strangs,
        using the cls._formatter
        """
        return self._formatter.format(self, *args, **kwargs)



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import os
import pathlib as pl
import pdb
import re
import signal
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging    = logmod.getLogger(__name__)
##-- end logging

env        : dict        = os.environ
PRE_COMMIT : Final[bool] = "PRE_COMMIT" in env
BREAK_HEADER : Final[str] = "\n---- Task Interrupted ---- "

##--| Body:

class SignalHandler:
    """ Install a breakpoint to run on (by default) SIGINT

    disables itself if PRE_COMMIT is in the environment.
    Can act as a context manager

    """

    def __init__(self):
        self._disabled = PRE_COMMIT

    @staticmethod
    def handle(signum, frame):
        breakpoint(header=BREAK_HEADER)
        SignalHandler.install()

    @staticmethod
    def install(sig=signal.SIGINT):
        logging.debug("Installing Basic Interrupt Handler for: %s", signal.strsignal(sig))
        signal.signal(sig, SignalHandler.handle)

    @staticmethod
    def uninstall(sig=signal.SIGINT):
        logging.debug("Uninstalling Basic Interrupt Handler for: %s", signal.strsignal(sig))
        signal.signal(sig, signal.SIG_DFL)

    def __enter__(self):
        if not self._disabled:
            SignalHandler.install()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if not self._disabled:
            SignalHandler.uninstall()
        return


class NullHandler:
    """ An interrupt handler that does nothing """

    @staticmethod
    def handle(signum, frame):
        return

    @staticmethod
    def install(sig=signal.SIGINT):
        logging.debug("Installing Null Interrupt handler for: %s", signal.strsignal(sig))
        # Install handler for Interrupt signal
        signal.signal(sig, NullHandler.handle)

    @staticmethod
    def uninstall(sig=signal.SIGINT):
        logging.debug("Uninstalling Null Interrupt handler for: %s", signal.strsignal(sig))
        signal.signal(sig, signal.SIG_DFL)

    def __enter__(self):
        if not self._disabled:
            NullHandler.install()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if not self._disabled:
            NullHandler.uninstall()
        return



#!/usr/bin/env python3
"""

"""
# mypy: disable-error-code="attr-defined"
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import functools as ftz
import itertools as itz
import logging as logmod
import logging.handlers as l_handlers
import os
import pathlib as pl
import re
import time
import types
import weakref
from sys import stderr, stdout
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv import Mixin, Proto
from jgdv._abstract.protocols import Buildable_p
from jgdv._abstract.pydantic_proto import ProtocolModelMeta
from jgdv.structs.chainguard import ChainGuard

# ##-- end 1st party imports

from . import _interface as API # noqa: N812
from .format import ColourFormatter, StripColourFormatter
from .filter import BlacklistFilter, WhitelistFilter

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Any
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError
if TYPE_CHECKING:
    import enum
    from jgdv import Maybe, RxStr
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable
    from ._interface import Handler, Formatter

##--|
from jgdv import Maybe
# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

env                  : dict             = os.environ # type: ignore
IS_PRE_COMMIT        : Final[bool]      = "PRE_COMMIT" in env
DEFAULT_FORMAT       : Final[str] = "{levelname:<8} : {message}"
DEFAULT_FILE_FORMAT  : Final[str] = "%Y-%m-%d::%H:%M.log"
DEFAULT_STYLE : Final[str] = "{"
##--|
class HandlerBuilder_m:
    """
    Loggerspec Mixin for building handlers
    """

    def _build_streamhandler(self) -> Handler:
        return logmod.StreamHandler(stdout)

    def _build_errorhandler(self) -> Handler:
        return logmod.StreamHandler(stderr)

    def _build_filehandler(self, path:pl.Path) -> Handler:
        return logmod.FileHandler(path, mode='w')

    def _build_rotatinghandler(self, path:pl.Path) -> Handler:
        handler = l_handlers.RotatingFileHandler(path, backupCount=API.MAX_FILES)
        handler.doRollover()
        return handler

    def _build_formatter(self, handler:Handler) -> Formatter:
        formatter : Formatter
        match self.colour:
            case _ if IS_PRE_COMMIT:
                # Always strip colour when in pre-commit
                formatter = StripColourFormatter(fmt=self.format, style=self.style)
            case _ if isinstance(handler, logmod.FileHandler|l_handlers.RotatingFileHandler):
                # Always strip colour when logging to a file
                formatter = StripColourFormatter(fmt=self.format, style=self.style)
            case False:
                formatter = StripColourFormatter(fmt=self.format, style=self.style)
            case str() | True:
                formatter = ColourFormatter(fmt=self.format, style=self.style)
                formatter.apply_colour_mapping(API.alt_log_colours)

        return formatter

    def _build_filters(self) -> list[Callable]:
        filters : list[Callable] = []
        if bool(self.allow):
            filters.append(WhitelistFilter(self.allow))
        if bool(self.filter):
            filters.append(BlacklistFilter(self.filter))

        return filters

    def _discriminate_handler(self, target:Maybe[str|pl.Path]) -> tuple[Maybe[Handler], Maybe[Formatter]]:
        handler, formatter = None, None

        match target:
            case "pass" | None:
                return None, None
            case "file":
                log_file_path      = self.logfile()
                handler            = self._build_filehandler(log_file_path)
            case "rotate":
                log_file_path      = self.logfile()
                handler            = self._build_rotatinghandler(log_file_path)
            case "stdout":
                handler   = self._build_streamhandler()
            case "stderr":
                handler = self._build_errorhandler()
            case _:
                msg = "Unknown logger spec target"
                raise ValueError(msg, target)

        formatter = self._build_formatter(handler)

        assert(handler   is not None)
        assert(formatter is not None)
        return handler, formatter

##--|

@Proto(Buildable_p)
class LoggerSpec(HandlerBuilder_m, BaseModel, metaclass=ProtocolModelMeta):
    """
      A Spec for toml defined logging control.
      Allows user to name a logger, set its level, format,
      filters, colour, and what (cli arg) verbosity it activates on,
      and what file it logs to.

      When 'apply' is called, it gets the logger,
      and sets any relevant settings on it.
    """
    ##--| classvars
    RootName                   : ClassVar[str]                   = "root"
    levels                     : ClassVar[type[enum.IntEnum]]    = API.LogLevel_e
    ##--| main
    name                       : str
    disabled                   : bool                            = False
    base                       : Maybe[str]                      = None
    level                      : str|int                         = logmod.WARNING
    format                     : str                             = DEFAULT_FORMAT
    filter                     : list[str]                       = []
    allow                      : list[str]                       = []
    colour                     : bool|str                        = False
    verbosity                  : int                             = 0
    target                     : list[str|pl.Path]               = [] # stdout | stderr | file
    filename_fmt               : str                             = DEFAULT_FILE_FORMAT
    propagate                  : bool                            = False
    clear_handlers             : bool                            = False
    style                      : str                             = DEFAULT_STYLE
    nested                     : list[LoggerSpec]                = []
    prefix                     : Maybe[str]                      = None
    ##--| internal
    _logger                    : Maybe[API.Logger]               = None
    _applied                   : bool                            = False

    @staticmethod
    def build(data:bool|list|dict, **kwargs:Any) -> LoggerSpec:  # noqa: ANN401
        """
          Build a single spec, or multiple logger specs targeting the same logger
        """
        match data:
            case LoggerSpec():
                return data
            case False:
                return LoggerSpec(disabled=True, **kwargs)
            case True:
                return LoggerSpec(**kwargs)
            case list():
                nested = []
                for x in data:
                    nested.append(LoggerSpec.build(x, **kwargs))
                return LoggerSpec(nested=nested, **kwargs)
            case ChainGuard():
                as_dict = data._table().copy()
                as_dict.update(kwargs)
                return LoggerSpec.model_validate(as_dict)
            case dict():
                as_dict = data.copy()
                as_dict.update(kwargs)
                return LoggerSpec.model_validate(as_dict)
            case _:
                msg = "Unknown data for logger spec"
                raise TypeError(msg, data)

    ##--| Validators

    @field_validator("level")
    def _validate_level(cls, val:str|int) -> int:  # noqa: N805
        match val:
            case str() if (lvl:=logmod.getLevelNamesMapping().get(val, None)) is not None:
                return lvl
            case str() if val in LoggerSpec.levels.__members__:
                return LoggerSpec.levels[val]
            case int():
                return val
            case _:
                raise ValueError(val)

    @field_validator("format")
    def _validate_format(cls, val:str) -> str:  # noqa: N805
        return val

    @field_validator("target", mode="before")
    def _validate_target(cls, val:list|str|pl.Path) -> list[str|pl.Path]:  # noqa: N805
        match val:
            case [*xs] if all(x in API.TARGETS for x in xs): # type: ignore
                return val
            case str() if val in API.TARGETS:
                return [val]
            case pl.Path():
                return [val]
            case None:
                return ["stdout"]
            case _:
                msg = "Unknown target value for LoggerSpec"
                raise ValueError(msg, val)

    @field_validator("style")
    def _validate_style(cls, val:str) -> str:  # noqa: N805
        match val:
            case "%" | "{" | "$":
                return val
            case _:
                msg = "API.Logger Style Needs to be in [{,%,$]"
                raise ValueError(msg, val)

    ##--| methods

    @ftz.cached_property
    def fullname(self) -> str:
        if self.base is None:
            return self.name
        return f"{self.base}.{self.name}"

    def apply(self, *, onto:Maybe[API.Logger]=None) -> API.Logger:  # noqa: PLR0912, PLR0915
        """ Apply this spec (and nested specs) to the relevant logger """
        logger : logmod.Logger
        match onto:
            case logmod.Logger() if self._applied:
                raise ValueError("Tried to apply logger when spec already has a logger", self.fullname)
            case logmod.Logger():
                logger = onto
            case None if self._applied:
                # already set up, just return it
                return self.get()
            case None:
                # not set up, get it and set it
                logger = self.get()

        handler_pairs : list[tuple[Maybe[Handler], Maybe[Formatter]]] = []
        logger.propagate = self.propagate
        logger.setLevel(logmod._nameToLevel.get("NOTSET", 0))
        if self.disabled:
            logger.disabled = True
            return logger

        match self.prefix:
            case str() if hasattr(logger, "set_prefixes"):
                logger.set_prefixes(self.prefix)
            case _:
                pass

        match self.colour:
            case str():
                logger.set_colour(self.colour)
            case _:
                pass

        match self.target:
            case _ if bool(self.nested):
                for subspec in self.nested:
                    subspec.apply(onto=logger)
                else:
                    return logger
            case []:
                handler_pairs.append(self._discriminate_handler(None))
            case [*xs]:
                handler_pairs += [self._discriminate_handler(x) for x in xs]
            case _:
                msg = "Unknown target value for LoggerSpec"
                raise ValueError(msg, self.target)

        log_filters = self._build_filters()
        for pair in handler_pairs:
            match pair:
                case None, _:
                    pass
                case hand, None:
                    hand.setLevel(self.level)
                    for fltr in log_filters:
                        hand.addFilter(fltr)
                    else:
                        logger.addHandler(hand)
                case hand, fmt:
                    hand.setLevel(self.level)
                    hand.setFormatter(fmt)
                    for fltr in log_filters:
                        hand.addFilter(fltr)
                    else:
                        logger.addHandler(hand)
                case _:
                    pass
        else:
            if not bool(logger.handlers):
                logger.setLevel(self.level)
                logger.propagate = True

            self._applied = True
            self._logger  = logger
            return logger

    def get(self) -> API.Logger:
        """ Get the logger  this spec controls """
        if self._logger is None:
            self._logger = logmod.getLogger(self.fullname)

        return self._logger

    def clear(self) -> None:
        """ Clear the handlers for the logger referenced """
        logger = self.get()
        handlers = logger.handlers[:]
        for h in handlers:
            logger.removeHandler(h)

        self._logger = None

    def logfile(self) -> pl.Path:
        log_dir  = pl.Path(".temp/logs")
        if not log_dir.exists():
            log_dir = pl.Path()

        filename = datetime.datetime.now().strftime(self.filename_fmt)  # noqa: DTZ005
        return log_dir / filename

    def set_level(self, level:int|str) -> None:
        match level:
            case str():
                level = logmod._nameToLevel.get(level, 0)
            case int():
                pass
        logger = self.get()
        logger.setLevel(level)
        for handler in logger.handlers:
            handler.setLevel(level)



#!/usr/bin/env python3
"""


"""
# ruff: noqa:

# Imports:
from __future__ import annotations

from jgdv.structs.strang.errors import StrangError

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
   from jgdv import Maybe
   from typing import Final
   from typing import ClassVar, Any, LiteralString
   from typing import Never, Self, Literal
   from typing import TypeGuard
   from collections.abc import Iterable, Iterator, Callable, Generator
   from collections.abc import Sequence, Mapping, MutableMapping, Hashable

# isort: on
# ##-- end types

# Vars:

# Body:

class LocationError(StrangError):
    """ A Task tried to access a location that didn't existing """
    general_msg = "Location Error:"

class LocationExpansionError(LocationError):
    """ When trying to resolve a location, something went wrong. """
    general_msg = "Expansion of Location hit max value:"

class DirAbsent(LocationError):
    """ In the course of startup verification, a directory was not found """
    general_msg = "Missing Directory:"



#!/usr/bin/env python3
"""

"""
# ruff: noqa: N801, ANN001, ANN002, ANN003
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from jgdv._abstract.protocols import SpecStruct_p
from jgdv.structs.strang import Strang, CodeReference
from .._interface import Key_p, NonKey_p

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
from collections.abc import Mapping
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, MutableMapping, Hashable

    from jgdv import Maybe, Rx, Ident, RxStr, Ctor, CHECKTYPE, FmtStr
    type LitFalse    = Literal[False]
    type InstructionList  = list[ExpInst_d]
    type InstructionAlts  = list[InstructionList]
    type ExpOpts          = dict
    type SourceBases      = list|Mapping|SpecStruct_p

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

##--| Error Messages
NestedFailure           : Final[str]  = "Nested ExpInst_d"
NoValueFailure          : Final[str]  = "ExpInst_d's must have a val"
UnexpectedData          : Final[str]  = "Unexpected kwargs given to ExpInst_d"

##--| Values
UNRESTRICTED_EXPANSION     : Final[int]                 = -1
NO_EXPANSIONS_PERMITTED    : Final[int]                 = 0

EXPANSION_CONVERT_MAPPING  : Final[dict[str,Callable]]  = {
    "p"                    : lambda x: pl.Path(x).expanduser().resolve(),
    "s"                    : str,
    "S"                    : Strang,
    "c"                    : CodeReference,
    "i"                    : int,
    "f"                    : float,
}
##--| Data

class ExpInst_d:
    """ The lightweight holder of expansion instructions, passed through the
    expander mixin.
    Uses slots to make it as lightweight as possible

    - fallback : the value to use if expansion fails
    - convert  : controls type coercion of expansion result
    - lift     : says to lift expanded values into keys themselves
    - literal  : signals the value needs no more expansion
    - rec      : the remaining recursive expansions available. -1 is unrestrained.
    - total_recs : tracks the number of expansions have occured

    """
    __slots__ = ("convert", "fallback", "lift", "literal", "rec", "total_recs", "value")
    value       : Any
    convert     : Maybe[str|bool]
    fallback    : Maybe[str]
    lift        : bool
    literal     : bool
    rec         : int
    total_recs  : int

    def __init__(self, **kwargs) -> None:
        self.value       = kwargs.pop("value")
        self.convert     = kwargs.pop("convert", None)
        self.fallback    = kwargs.pop("fallback", None)
        self.lift        = kwargs.pop("lift", False)
        self.literal     = kwargs.pop("literal", False)
        self.rec         = kwargs.pop("rec", None) or UNRESTRICTED_EXPANSION
        self.total_recs  = kwargs.pop("total_recs", 0)

        self.process_value()
        if bool(kwargs):
            raise ValueError(UnexpectedData, kwargs)

    def __repr__(self) -> str:
        lit  = "(Lit)" if self.literal else ""
        return f"<ExpInst_d:{lit} {self.value!r} / {self.fallback!r} (R:{self.rec},L:{self.lift},C:{self.convert})>"

    def process_value(self) -> None:
        match self.value:
            case ExpInst_d() as val:
                raise TypeError(NestedFailure, val)
            case None:
                 raise ValueError(NoValueFailure)
            case Key_p():
                pass

class SourceChain_d:
    """ The core logic to lookup a key from a sequence of sources

    | Doesn't perform repeated expansions.
    | Tries sources in order.

    TODO replace this with collections.ChainMap ?
    """
    __slots__ = ("lifter", "sources")
    sources  : list[Mapping|list]
    lifter   : type[Key_p]

    def __init__(self, *args:Maybe[SourceBases|SourceChain_d], lifter=type[Key_p]) -> None:
        self.sources = []
        for base in args:
            match base:
                case None:
                    pass
                case SourceChain_d():
                    self.sources += base.sources
                case dict() | collections.ChainMap() | list():
                    self.sources.append(base)
                case Mapping():
                    self.sources.append(base)
                case SpecStruct_p():
                    self.sources.append(base.params)
                case x:
                    raise TypeError(type(x))
        self.lifter   = lifter

    def extend(self, *args:SourceBases) -> Self:
        extension = SourceChain_d(*args)
        self.sources += extension.sources
        return self

    def get(self, key:str, fallback:Maybe=None) -> Maybe:
        """ Get a key's value from an ordered sequence of potential sources.

        | Try to get {key} then {key\\_} in order of sources passed in.
        """
        replacement  : Maybe  = fallback
        for lookup in self.sources:
            match lookup:
                case None | []:
                    continue
                case list():
                    replacement = lookup.pop()
                case _ if hasattr(lookup, "get"):
                    if key not in lookup:
                        continue
                    replacement = lookup.get(key, fallback)
                case SpecStruct_p():
                    params      = lookup.params
                    replacement = params.get(key, fallback)
                case _:
                    msg = "Unknown Type in get"
                    raise TypeError(msg, key, lookup)

            if replacement is not fallback:
                return replacement
        else:
            return fallback



    def lookup(self, target:list[ExpInst_d]) -> Maybe[ExpInst_d]:
        """ Look up alternatives

        | pass through DKeys and (DKey, ..) for recursion
        | lift (str(), True, fallback)
        | don't lift (str(), False, fallback)

        """
        x : Any
        for spec in target:
            match spec:
                case ExpInst_d(value=Key_p()):
                    return spec
                case ExpInst_d(literal=True):
                    return spec
                case ExpInst_d(value=str() as key, lift=lift, fallback=fallback):
                    pass
                case x:
                    msg = "Unrecognized lookup spec"
                    raise TypeError(msg, x)

            match self.get(key):
                case None:
                    pass
                case x if lift:
                    logging.debug("Lifting Result to Key: %r", x)
                    match self.lifter(x, implicit=True, fallback=fallback): # type: ignore[call-arg]
                        case NonKey_p() as y:
                            return ExpInst_d(value=y, rec=0, litreal=True)
                        case Key_p() as y:
                            return ExpInst_d(value=y, fallback=fallback, lift=False)
                        case x:
                            raise TypeError(type(x))
                case x:
                    return ExpInst_d(value=x, fallback=fallback)
        else:
            return None
##--| Protocols

class Expander_p[T](Protocol):

    def set_ctor(self, ctor:Ctor[T]) -> None: ...

    def redirect(self, source:T, *sources:dict, **kwargs:Any) -> list[Maybe[ExpInst_d]]:  ...  # noqa: ANN401

    def expand(self, source:T, *sources:dict, **kwargs:Any) -> Maybe[ExpInst_d]:  ...  # noqa: ANN401

    def extra_sources(self, source:T) -> SourceChain_d: ...
    def coerce_result(self, inst:ExpInst_d, opts:ExpOpts, *, source:Key_p) -> Maybe[ExpInst_d]: ...

class ExpansionHooks_p(Protocol):

    def exp_extra_sources_h(self, current:SourceChain_d) -> SourceChain_d: ...

    def exp_generate_alternatives_h(self, sources:SourceChain_d, opts:ExpOpts) -> InstructionAlts: ...

    def exp_configure_recursion_h(self, insts:InstructionAlts, sources:SourceChain_d, opts:ExpOpts) -> Maybe[InstructionList]: ...

    def exp_flatten_h(self, insts:InstructionList, opts:dict) -> Maybe[ExpInst_d]: ...

    def exp_coerce_h(self, inst:ExpInst_d, opts:dict) -> Maybe[ExpInst_d]: ...

    def exp_final_h(self, inst:ExpInst_d, opts:dict) -> Maybe[LitFalse|ExpInst_d]: ...

    def exp_check_result_h(self, inst:ExpInst_d, opts:dict) -> None: ...

class Expandable_p(Protocol):
    """ An expandable, like a DKey,
    uses these hooks to customise the expansion
    """

    def expand(self, *sources, **kwargs) -> Maybe: ...



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ARG002

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
# ##-- end stdlib imports

from collections import ChainMap
from jgdv import Proto, Mixin
from jgdv.mixins.annotate import SubAlias_m
from jgdv.structs.strang import Strang
from ._util import _interface as ExpAPI # noqa: N812
from ._util.expander import DKeyExpander
from .processor import DKeyProcessor

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol
from . import _interface as API # noqa: N812

if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, M_
    from ._util._interface import Expander_p


# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
CLASS_GETITEM_K : Final[str] = "__class_getitem__"
type DP_TYPE = API.DKeyMarkAbstract_e | type
# Body:

@Proto(API.Key_p, check=False, mod_mro=False)
class DKey[**K](Strang, fresh_registry=True):
    """ A facade for DKeys and variants.
      Implements __new__ to create the correct key type, from a string, dynamically.

      kwargs:
      explicit = insists that keys in the string are wrapped in braces '{akey} {anotherkey}'.
      mark     = pre-register expansion parameters / type etc
      check    = dictate a type that expanding this key must match
      fparams  = str formatting instructions for the key

      Eg:
      DKey('blah')
      -> SingleDKey('blah')
      -> SingleDKey('blah').format('w')
      -> '{blah}'
      -> [toml] aValue = '{blah}'

      Because cls.__new__ calls __init__ automatically for return values of type cls,
      DKey is the factory, but all DKeys are subclasses of DKeyBase,
      to allow control over __init__.

      Base class for implementing actual DKeys.

      init takes kwargs:
      fmt, mark, check, ctor, help, fallback, max_exp

      on class definition, can register a 'mark', 'multi', and a conversion parameter str
    """
    __slots__                                             = ("data",)
    __match_args                                          = ()
    _annotate_to    : ClassVar[str]                       = "dkey_mark"
    _processor      : ClassVar                            = DKeyProcessor()
    _sections       : ClassVar                            = API.DKEY_SECTIONS
    _expander       : ClassVar[Expander_p]                = cast("Expander_p", DKeyExpander())
    _typevar        : ClassVar                            = None
    _extra_kwargs   : ClassVar[set[str]]                  = set()
    _extra_sources  : ClassVar[list[ExpAPI.SourceBases]]  = []
    Marks           : ClassVar[API.DKeyMarkAbstract_e]    = API.DKeyMark_e  # type: ignore[assignment]
    data            : API.DKey_d

    ##--| Class Utils

    @final
    @staticmethod
    def MarkOf[T](target:T|type[T]) -> API.KeyMark|tuple[API.KeyMark, ...]: # noqa: N802
        """ Get the mark of the key type or instance """
        if not hasattr(target, "cls_annotation"):
            return ()
        match target.cls_annotation(): # type: ignore[union-attr]
            case None:
                return ()
            case [x]:
                return x
            case xs:
                return xs

    @classmethod
    def add_sources(cls, *sources:dict) -> None:
        """ register additional sources that are always included in expansion """
        cls._extra_sources += sources


    def __init_subclass__(cls, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        super().__init_subclass__(*args, annotation=kwargs.pop("mark", None), **kwargs)
        cls._expander.set_ctor(DKey) # type: ignore[arg-type]
        cls._processor.register_convert_param(cls, kwargs.pop("convert", None))


    ##--| Class Main

    def __init__(self, *args:Any, **kwargs:Any) -> None:  # noqa: ANN401
        assert(not self.endswith(API.INDIRECT_SUFFIX)), self[:]
        super().__init__(*args, **kwargs)
        self.data = API.DKey_d(**kwargs)

    def __call__(self, *args:Any, **kwargs:Any) -> Any:  # noqa: ANN401
        """ call expand on the key.
        Args and kwargs are passed verbatim to expand()
        """
        return self.expand(*args, **kwargs)

    def __eq__(self, other:object) -> bool:
        match other:
            case DKey() | str():
                return str.__eq__(self, other)
            case _:
                return NotImplemented


    def __hash__(self) -> int:
        return hash(self[:])

    def __format__(self, spec:str) -> str:
        """
          Extends standard string format spec language:
            [[fill]align][sign][z][#][0][width][grouping_option][. precision][type]
            (https://docs.python.org/3/library/string.html#format-specification-mini-language)

          Using the # alt form to declare keys are wrapped.
          eg: for key = DKey('test'), ikey = DKey('test_')
          f'{key}'   -> 'test'
          f'{key:w}' -> '{test}'
          f'{key:i}  ->  'test_'
          f'{key:wi} -> '{test_}'

          f'{ikey:d} -> 'test'

        """
        result = self[:]
        if not bool(spec):
            return result

        rem, wrap, direct = self._processor.consume_format_params(spec) # type: ignore

        # format
        if not direct:
            result = f"{result}{API.INDIRECT_SUFFIX}"

        if wrap:
            result = "".join(["{", result, "}"])  # noqa: FLY002

        return result
    ##--| Utils

    def var_name(self) -> str:
        """ When testing the dkey for its inclusion in a decorated functions signature,
        this gives the 'named' val if its not None, otherwise the str of the key
        """
        return self.data.name or str(self)

    def expand(self, *args:Any, **kwargs:Any) -> Maybe:  # noqa: ANN401
        kwargs.setdefault("limit", self.data.max_expansions)
        assert(isinstance(self, API.Key_p))
        match self._expander.expand(self, *args, **kwargs):
            case ExpAPI.ExpInst_d(value=val, literal=True):
                return val
            case _:
                return None

    def redirect(self, *args:Any, **kwargs:Any) -> list[API.Key_p]:  # noqa: ANN401
        assert(isinstance(self, API.Key_p))
        result = [DKey(x.value) for x in self._expander.redirect(self, *args, **kwargs) if x is not None]
        return result


    ##--| expansion hooks
    def to_exp_inst(self, *, indirect:bool=False, **kwargs:Any) -> ExpAPI.ExpInst_d:  # noqa: ANN401
        """ create a basic expinst """
        val = f"{self:i}" if indirect else f"{self:d}"
        return ExpAPI.ExpInst_d(value=val,
                                convert=self.data.convert,
                                rec=self.data.max_expansions,
                                **kwargs,
                                )

    def exp_extra_sources_h(self, current:ExpAPI.SourceChain_d) -> ExpAPI.SourceChain_d:
        match self._extra_sources:
            case [*xs]:
                return current.extend(*xs)
            case x:
                raise TypeError(type(x))



#!/usr/bin/env python3

"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

##-- end builtin imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from dataclasses import InitVar, dataclass, field

if TYPE_CHECKING:
    from jgdv import Maybe, RxStr
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@dataclass
class RxMatcher(str):
    """ https://martinheinz.dev/blog/78 """
    string : str
    match  : Maybe[re.Match] = None

    def __eq__(self, pattern:RxStr) -> bool:
        self.match = re.search(pattern, self.string)
        return self.match is not None

    def __getitem__(self, group:int) -> str: #type:ignore[override]
        return self.match[group]



#!/usr/bin/env python3
"""

"""
# ruff: noqa: B011
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import warnings
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest

# ##-- end 3rd party imports

# ##-- 1st party imports
from .. import _interface as API
from .._core import (
    Decorator,
    IdempotentDec,
    MonotonicDec,
)

from jgdv.decorators.mixin import Mixin
# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

logging = logmod.root

##-- mixins

class Simple_m:

    def blah(self):
        return 2

    def bloo(self):
        return 4

class Second_m:

    def aweg(self):
        return super().bloo()

##-- end mixins

##-- super chain

class ChainBase:

    def vals(self) -> list:
        return ["ChainBase"]

class m1:

    def vals(self) -> list:
        return ["m1", *super().vals()]

class m2:

    def vals(self) -> list:
        return ["m2", *super().vals()]

class m3:

    def vals(self) -> list:
        return ["m3", *super().vals()]

##-- end super chain

class TestMixinDecorator:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):

        @Mixin(None, Simple_m)
        class Example:

            def bloo(self):
                return 10

        obj = Example()
        assert(obj.blah() == 2)
        assert(obj.bloo() == 10)

    def test_two_mixins(self):

        @Mixin(None, Second_m, Simple_m)
        class Example:

            def bloo(self):
                return 10

        obj = Example()
        assert(obj.blah() == 2)
        assert(obj.bloo() == 10)
        # Aweg->super()->Simple_m.bloo
        assert(obj.aweg() == 4)

    def test_append_mixin(self):

        @Mixin(Second_m, None, Simple_m)
        class Example:
            val : ClassVar[int] = 25

            def bloo(self):
                return 10

        obj = Example()
        assert(obj.blah() == 2)
        assert(obj.bloo() == 10)
        # Aweg->super()->Example.bloo
        assert(obj.aweg() == 10)
        assert(Example.val == 25)

    def test_super_chain_traditional(self):

        class ChainTop(m1, m2, m3, ChainBase):

            def vals(self) -> list:
                return ["ChainTop", *super().vals()]

        obj      = ChainTop()
        real     = obj.vals()
        expected = ["ChainTop", "m1", "m2", "m3", "ChainBase"]
        for x,y in zip(real, expected, strict=True):
            assert(x == y), real

    def test_pre_super_chain_decorated(self):

        @Mixin(m1, m2, m3, None)
        class PreChainTop(ChainBase):

            def vals(self) -> list:
                return ["ChainTop", *super().vals()]

        obj     = PreChainTop()
        real     = obj.vals()
        expected = ["m1", "m2", "m3", "ChainTop", "ChainBase"]
        for x,y in zip(real, expected, strict=True):
            assert(x == y), real

    def test_post_super_chain_decorated(self):

        @Mixin(None, m1, m2, m3)
        class PostChainTop(ChainBase):

            def vals(self) -> list:
                return ["ChainTop", *super().vals()]

        obj      = PostChainTop()
        real     = obj.vals()
        expected = ["ChainTop", "m1", "m2", "m3", "ChainBase"]
        for x,y in zip(real, expected, strict=True):
            assert(x == y), real

    def test_spit_super_chain_decorated(self):

        @Mixin(m1, None, m2, m3)
        class PostChainTop(ChainBase):

            def vals(self) -> list:
                return ["ChainTop", *super().vals()]

        obj      = PostChainTop()
        real     = obj.vals()
        expected = ["m1", "ChainTop", "m2", "m3", "ChainBase"]
        for x,y in zip(real, expected, strict=True):
            assert(x == y), real


    def test_multi_dec_super_chain(self):

        @Mixin(m1, None)
        @Mixin(m2, m3)
        class PostChainTop(ChainBase):

            def vals(self) -> list:
                return ["ChainTop", *super().vals()]

        obj      = PostChainTop()
        real     = obj.vals()
        expected = ["m1", "ChainTop", "m2", "m3", "ChainBase"]
        for x,y in zip(real, expected, strict=True):
            assert(x == y), real

    def test_duplicate_errors(self):

        with pytest.raises(TypeError):

            @Mixin(m1, m1)
            class PostChainTop(ChainBase):
                pass



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN202, B011
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, ClassVar, Generic, TypeAlias, TypeVar, cast, Final)
import warnings

import pytest
logging = logmod.root

from jgdv import identity_fn
from jgdv.structs.strang import Strang
from jgdv.structs.strang import _interface as API  # noqa: N812
from jgdv.structs.strang.code_ref import CodeReference

EX_STR     : Final[str]  = "fn::jgdv:identity_fn"
NO_PREFIX  : Final[str]  = "jgdv:identity_fn"

class TestCodeReference:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        ref = CodeReference(EX_STR)
        assert(isinstance(ref, CodeReference))
        assert(ref == EX_STR)

    def test_annotated(self):
        ref = CodeReference[bool]
        assert(issubclass(ref, CodeReference))
        assert(ref.cls_annotation() == (bool,))
        inst = ref(EX_STR)
        assert(isinstance(inst, CodeReference))

    def test_with_no_prefix(self):
        ref = CodeReference(NO_PREFIX)
        assert(isinstance(ref, CodeReference))

    def test_instance(self):
        ref = CodeReference(EX_STR)
        assert(isinstance(ref, CodeReference))

    def test_with_value(self):
        ref = CodeReference(EX_STR, value=int)
        assert(isinstance(ref, CodeReference))

    def test_str(self):
        ref = CodeReference(EX_STR)
        assert(str(ref) == EX_STR)

    def test_repr(self):
        ref = CodeReference(EX_STR)
        assert(repr(ref) == f"<CodeReference: {EX_STR}>")

    def test_module(self):
        ref = CodeReference(EX_STR)
        assert(ref[1,0] == "jgdv")
        assert(ref.module == "jgdv")

    def test_value(self):
        ref = CodeReference(EX_STR)
        assert(ref[2,0] == "identity_fn")
        assert(ref.value == "identity_fn")

    @pytest.mark.xfail
    def test_from_fn(self):

        def simple_fn():
            return "blah"

        ref = CodeReference(simple_fn)
        assert(ref() is simple_fn)

class TestCodeReference_Importing:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_import(self):
        ref      = CodeReference(EX_STR)
        match ref(check=False):
            case Exception() as x:
                assert(False), x
            case x:
                assert(callable(x))
                assert(x == identity_fn)

    def test_import_module_fail(self):
        ref = CodeReference("cls::jgdv.taskSSSSS.base_task:DootTask")
        match ref():
            case ImportError():
                assert(True)
            case x:
                assert(False), x

    def test_import_non_existent_class_fail(self):
        ref = CodeReference("cls::jgdv.structs.strang:DootTaskSSSSSS")
        match ref():
            case ImportError():
                assert(True)
            case _:
                assert(False)

    def test_import_non_class_fail(self):
        ref = CodeReference("cls::jgdv.structs.strang.strang:GEN_K")
        match ref():
            case ImportError():
                assert(True)
            case _:
                assert(False)

    def test_import_non_callable(self):
        ref = CodeReference("fn::jgdv.structs.strang.strang:GEN_K")
        match ref():
            case ImportError():
                assert(True)
            case _:
                assert(False)

    def test_import_value(self):
        ref = CodeReference("val::jgdv.structs.strang._interface:GEN_K")
        assert(ref[0,:] == "val")
        assert(ref[1,:] == "jgdv.structs.strang._interface")
        assert(ref[2,:] == "GEN_K")
        match ref(check=False):
            case str() as x:
                assert(x == API.GEN_K)
            case ImportError():
                assert(False)
            case _:
                assert(True)

class TestCodeReference_TypeCheck:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic_typecheck(self):
        ref = CodeReference[Strang]("cls::jgdv.structs.strang:Strang")
        match ref():
            case type() as x if x == Strang:
                assert(True)
            case x:
                assert(False), x

    def test_union_typecheck(self):
        ref = CodeReference[int|Strang]("cls::jgdv.structs.strang:Strang")
        assert(ref.expects_type() == int|Strang)
        match ref():
            case type() as x if x == Strang:
                assert(True)
            case x:
                assert(False), x

    def test_typecheck_fail(self):
        ref = CodeReference[bool]("cls::jgdv.structs.strang:Strang")
        assert(ref.expects_type() is bool)
        match ref():
            case ImportError():
                assert(True)
            case x:
                assert(False), x

    def test_union_typecheck_fail(self):
        ref = CodeReference[bool|int]("cls::jgdv.structs.strang:Strang")
        assert(ref.expects_type() == bool|int)
        match ref():
            case ImportError():
                assert(True)
            case x:
                assert(False), x

    def test_explicit_typecheck(self):
        ref = CodeReference("cls::jgdv.structs.strang:Strang")
        assert(ref.expects_type() is None)
        match ref(check=Strang):
            case type():
                assert(True)
            case x:
                assert(False), x

    def test_explicit_union_typecheck(self):
        ref = CodeReference("cls::jgdv.structs.strang:Strang")
        assert(ref.expects_type() is None)
        match ref(check=int|Strang):
            case type():
                assert(True)
            case x:
                assert(False), x

class TestCodeReference_Annotate:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_no_annotation(self):
        basic = CodeReference(EX_STR)
        assert(basic._check is None)
        assert(basic.expects_type() is None)


    def test_annotation_doesnt_override_super(self):
        basic = CodeReference(EX_STR)
        assert(basic._check is None)
        assert(basic.expects_type() is None)
        other = CodeReference[int](EX_STR)
        assert(basic.expects_type() != other.expects_type())


    def test_annotation_doesnt_register(self):
        basic = CodeReference
        anno1 = CodeReference[int]
        anno2 = CodeReference[int|str]
        anno3 = CodeReference[str|Strang|float]

        for x in [basic, anno1, anno2, anno3]:
            assert(x not in CodeReference._registry)

    def test_simple(self):
        basic = CodeReference[int](EX_STR)
        assert(basic._check is int)
        assert(basic.expects_type() is int)

    def test_union(self):
        basic = CodeReference[int|str](EX_STR)
        assert(basic._check == int|str)
        assert(basic.expects_type() == int|str)



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN202, B011, PLR2004, ANN001
from __future__ import annotations

import uuid
import logging as logmod
import pathlib as pl
from types import GenericAlias
from typing import (Any, Annotated, ClassVar, Generic, TypeAlias,
                    TypeVar, cast)
from re import Match
from collections.abc import Callable, Iterable, Iterator, Mapping, MutableMapping, Sequence
import warnings
import pytest
from random import randint

from .. import _interface as API  # noqa: N812
from ..errors import StrangError
from ..strang import Strang
from ..processor import StrangBasicProcessor

##--|
logging  = logmod.root
UUID_STR = str(uuid.uuid1())

##--|

class TestStrang_Base:
    """ Ensure basic functionality of structured names,
    but ensuring StrName is a str.
    """

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_type_subclass(self):
        assert(issubclass(Strang, str))

    def test_basic_ctor(self):
        ing = "head.a::tail.b"
        obj = Strang(ing)
        assert(obj is not ing)
        assert(isinstance(obj, Strang))
        assert(isinstance(obj, API.Strang_p))
        assert(isinstance(obj, str))
        assert(not hasattr(obj, "__dict__"))
        assert(not obj.uuid())
        assert(obj.shape == (2, 2))

    def test_ctor_with_multi_args(self):
        ing = "head.a::tail"
        args = ["a","b","c"]
        obj = Strang(ing, *args)
        assert(obj is not ing)
        assert(obj == "head.a::tail.a.b.c")

    def test_typing(self):
        obj = Strang("head::tail")
        assert(isinstance(obj, Strang))
        assert(isinstance(obj, str))
        assert(str in Strang.mro())

    def test_needs_separator(self):
        with pytest.raises(StrangError):
            Strang("head|tail")

    def test_shape(self):
        obj = Strang("head.a.b::tail.c.d.blah.bloo")
        assert(obj.shape == (3,5))

class TestStrang_GetItem:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_equiv_to_str(self):
        ing = "group.blah.awef::a.b.c"
        ang = Strang(ing)
        assert(ang is not ing)
        assert(ang == ing)
        assert(ang[0] == ing[0])
        assert(ang[:-1] == ing[:-1])
        assert(ang[2:6] == ing[2:6])
        assert(ang[:] == ing[:])
        assert(ang[:] is not ing[:])

    def test_section_word(self):
        val = Strang("a.b.c::d.cognate.f")
        assert(val[0,0]   == "a")
        assert(val[0,1]   == "b")
        assert(val[0,-1]  == "c")
        assert(val[1,0]   == "d")
        assert(val[1,1]   == "cognate")
        assert(val[1,-1]  == "f")

    def test_section_word_slice(self):
        val = Strang("a.b.c::d.e.f.g")
        assert(val[0,:]   == "a.b.c")
        assert(val[0,:-1] == "a.b")
        assert(val[1,0::2] == "d.f")
        assert(val[1,:0:-2] == "g.e")

    def test_section_word_by_name(self):
        val = Strang("a.b.c::d.cognate.f")
        assert(val['head',0]   == "a")
        assert(val['head',1]   == "b")
        assert(val['head',-1]  == "c")
        assert(val['body',0]   == "d")
        assert(val['body',1]   == "cognate")
        assert(val['body',-1]  == "f")

    def test_section_slice(self):
        val = Strang("a.b.c::d.e.f")
        match val.data.sections[0]:
            case slice(start=x,stop=y):
                assert(x == 0)
                assert(y == 5)
            case x:
                 assert(False), x

        assert(val[0,:] == "a.b.c")
        assert(val[1,:] == "d.e.f")

    def test_section_slice_by_name(self):
        val = Strang("a.b.c::d.e.f")
        match val.data.sections[0]:
            case slice(start=x,stop=y):
                assert(x == 0)
                assert(y == 5)
            case x:
                 assert(False), x

        assert(val['head',:] == "a.b.c")
        assert(val['body',:] == "d.e.f")
        assert(val['head'] == "a.b.c")
        assert(val['body'] == "d.e.f")

    def test_multi_slice(self):
        """
        a.b.c::d.e.f
        [:,:,:-1] -> a.b.c::d.e
        """
        val = Strang("a.b.c::d.e.f")
        assert(val[:,:,:-1] == "a.b.c::d.e")
        assert(val[:,1:,:] == "b.c::d.e.f")
        assert(val[:,2,1] == "c::e")

    def test_body_mark(self):
        val = Strang("group.blah.awef::a.$head$.c")
        assert(val[1, 1] == "$head$")
        assert(val.get(1,1) is val.section(1).marks.head)
        assert(Strang.section(1).marks.head in val)

    def test_head_mark(self):
        val = Strang("group.blah.$basic$::a..$head$")
        assert(val[0, -1] == "$basic$")
        assert(val.get(0, -1) is val.section(0).marks.basic)
        assert(Strang.section(0).marks.basic in val)

    def test_multi_slices(self):
        """
        [:,:] gets the strang, but uses Strang.get instead of str.__getitem__
        """
        ing = "group.blah.$basic$::a..$head$"
        ang = Strang(ing)
        assert(ang[:,:] == ing)

    def test_multi_slices_not_all_sections(self):
        """
        [:1,:] gets the strang, but uses Strang.get instead of str.__getitem__
        """
        ing = "group.blah.$basic$::a..$head$"
        ang = Strang(ing)
        assert(ang[:1,:] == "group.blah.$basic$::")
        assert(ang[1:,:] == "a..$head$")

    def test_section_slice_error_on_negative(self):
        """
        [:1,:] gets the strang, but uses Strang.get instead of str.__getitem__
        """
        ing = "group.blah.$basic$::a..$head$"
        ang = Strang(ing)
        with pytest.raises(ValueError):
            ang[:-1,:]

    def test_multi_slices_with_implicit_uuid(self):
        """
        [:,:] gets the strang, but expands the values to give uuid vals as well
        """
        ing       = "group.blah.$basic$::a.b.c.<uuid>"
        ang       = Strang(ing)
        uuid_ing  = f"group.blah.$basic$::a.b.c.<uuid:{ang.get(1,-1)}>"
        ##--|
        assert(ang[:] == ing)
        assert(ang[:,:] == uuid_ing)

    def test_multi_slices_with_explicit_uuid(self):
        """
        [:,:] gets the strang, but expands the values to give uuid vals as well
        because the underlying string is explicit, [:] also shows the uuid value
        """
        uuid_obj         = uuid.uuid1()
        ing              = f"group.blah.$basic$..<uuid:{uuid_obj}>::a.b.c"
        ang              = Strang(ing)
        assert(ang.get(0,-1) == uuid_obj)
        assert(ang[:,:]  == ing)

class TestStrang_GetAttr:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_head(self):
        obj = Strang("head.a.b::tail.c.d")
        assert(obj.head == "head.a.b")

    def test_body(self):
        obj = Strang("head.a.b::tail.c.d")
        assert(obj.body == "tail.c.d")

    def test_missing(self):
        obj = Strang("head.a.b::tail.c.d")
        with pytest.raises(AttributeError):
            assert(obj.tail)

class TestStrang_Get:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_str(self):
        val = Strang("group.blah.awef::a.blah.2")
        match val.get(1,1), val[1,1]:
            case "blah", "blah":
                assert(True)
            case x:
                 assert(False), x

    def test_int(self):
        val = Strang("group.blah.awef::a.blah.<int:3>")
        match val.get(1,-1), val[1,-1]:
            case 3, "<int>":
                assert(True)
            case x:
                assert(False), (type(x), x)

    def test_uuid(self):
        val = Strang("group.blah.awef::a.<uuid>")
        assert(not val.uuid())
        match val.get(1, -1), val[1,-1]:
            case uuid.UUID() as x, "<uuid>":
                assert(True)
            case x:
                assert(False), x

    def test_uuid_repeat(self):
        val = Strang("group.blah.awef.<uuid>::a.ab.c")
        assert(val.get(0,-1) is val.get(0,-1))

class TestStrang_Words:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_get_words(self):
        val = Strang("group.blah.awef.<uuid>::a.b")
        assert(not val.uuid())
        match list(val.words(0)), val[0,:]:
            case [*xs], "group.blah.awef.<uuid>":
                for x,y in zip(xs, ["group","blah", "awef", val.get(0,-1)], strict=True):
                    assert(x == y)
                else:
                    assert(True)
            case x:
                 assert(False), x

    def test_get_words_with_case(self):
        val = Strang("group.blah.awef::a.b.<uuid>")
        assert(not val.uuid())
        match list(val.words(1, case=True)), val[1,:]:
            case [*xs], "a.b.<uuid>":
                for x,y in zip(xs, ["a", ".", "b", ".", val.get(1,-1)], strict=True):
                    assert(x == y)
                else:
                    assert(True)
            case x:
                 assert(False), x

    def test_words_with_selection(self):
        val    = Strang("group.blah.awef::a.b.c.d.e.f")
        words  = list(val.words(1, case=True, select=slice(0, 3)))
        match words, val[1,:3]:
            case [*xs], "a.b.c" as targ:
                assert("".join(xs) == targ)
                for x,y in zip(xs, ["a",".","b",".","c"],
                               strict=True):
                    assert(x == y)
                else:
                    assert(True)
            case x:
                 assert(False), x

class TestStrang_Iter:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_iter(self):
        val = Strang("group.blah.awef::a.b.c")
        for x,y in zip(val, ["group", "blah","awef", "a", "b", "c"],
                       strict=True):
            assert(x == y)

    def test_iter_uuid(self):
        val = Strang("group.blah.awef::a.b.c.<uuid>")
        assert(not val.uuid())
        for x,y in zip(val, ["group", "blah", "awef", "a", "b","c", val.get(1,-1)],
                       strict=False):
            assert(x == y)

class TestStrang_Shape:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_shape(self):
        obj = Strang("a.b.c::d.e.f.g.h")
        assert(obj.shape == (3,5))

class TestStrang_Index:

    def test_index(self):
        ing = "a.b.c::c.e.f"
        ang = Strang(ing)
        assert(ing.index("b") == ang.index("b"))

    def test_index_mark(self):
        ing = "a.b.c::d.e.$head$.f"
        ang = Strang(ing)
        assert(ang.index("$head$") == ang.index(Strang.section(1).marks.head))

    def test_index_mark_empty_str(self):
        """ find('') is useless,
        but find(mark.empty) can
        """
        ing = "a.b.c::d.e..f"
        ang = Strang(ing)
        with pytest.raises(ValueError):
            ang.index("")

        index1 = ang.index(ang.section(1).marks.empty)
        index2 = ang.index(1,2)
        assert(index1 == index2)

    def test_index_slice(self):
        ing = "a.b.c::d.blah.f"
        ang = Strang(ing)
        assert(ang.index(1,1) == ing.index("blah"))

    def test_index_word_slice(self):
        ing = "a.b.c::d.blah.f"
        ang = Strang(ing)
        idx1 = ang.index('body',1)
        idx2 = ing.index("blah")
        assert(idx1 == idx2)

    def test_index_fail(self):
        ing = "a.b.c::d.e.f"
        ang = Strang(ing)
        with pytest.raises(ValueError):
            ang.index("g")

    def test_index_mark_fail(self):
        ing = "a.b.c::d.e.f"
        ang = Strang(ing)
        with pytest.raises(ValueError):
            ang.index(Strang.section(1).marks.head)

    def test_rindex(self):
        ing = "a.b.c::c.e.f"
        ang = Strang(ing)
        assert(ing.rindex("c") == ang.rindex("c"))

    def test_rindex_mark_empty_str(self):
        """ find('') is useless,
        but find(mark.empty) can
        """
        ing = "a.b.c::d.e..f"
        ang = Strang(ing)
        with pytest.raises(ValueError):
            ang.rindex("")

        rindex1 = ang.rindex(ang.section(1).marks.empty)
        rindex2 = ang.rindex(1,2)
        assert(rindex1 == rindex2)

    def test_rindex_fail(self):
        ing = "a.b.c::c.e.f"
        ang = Strang(ing)
        with pytest.raises(ValueError):
            ang.rindex("g")

class TestStrang_EQ:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_eq_to_str(self):
        obj = Strang("head::tail.a.b.c")
        other = "head::tail.a.b.c"
        assert(obj == other)

    def test_eq_to_strang(self):
        obj = Strang("head::tail.a.b.c")
        other = Strang("head::tail.a.b.c")
        assert(obj == other)

    def test_not_eq_to_strang(self):
        obj = Strang("head::tail.a.b.c")
        other = Strang("head::tail.a.b.c.d")
        assert(obj != other)

    def test_not_eq_to_strang_group(self):
        obj = Strang("head::tail.a.b.c")
        other = Strang("head.blah::tail.a.b.c")
        assert(obj != other)

    def test_not_eq_uuids(self):
        obj   = Strang("head::tail.a.<uuid>")
        other = Strang("head::tail.a.<uuid>")
        match obj.get(1,-1), other.get(1,-1):
            case uuid.UUID() as x, uuid.UUID() as y if x != y:
                assert(obj.__eq__(other) is (obj == other))
                assert(obj != other)
            case x:
                assert(False), x

    def test_not_eq_to_str(self):
        obj    = Strang("head::tail.a.b.c")
        other  = "tail.a.b.c.d"
        assert(obj != other)

class TestStrang_Hash:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_hash(self):
        obj  = Strang("head::tail.a.b.c")
        obj2 = Strang("head::tail.a.b.c")
        assert(hash(obj) == hash(obj2))

    def test_hash_same_as_str_hash(self):
        obj = Strang("head::tail.a.b.c")
        assert(hash(obj) == str.__hash__(obj))
        assert(hash(obj) == hash(str(obj)))

    def test_hash_spy(self, mocker):
        hash_spy = mocker.spy(Strang, "__hash__")
        obj = Strang("head::tail.a.b.c")
        hash(obj)
        hash_spy.assert_called()

    def test_hash_fail(self):
        obj  = Strang("head::tail.a.b.c")
        obj2 = Strang("head::tail.a.b.d")
        assert(hash(obj) != hash(obj2))

    def test_different_uuids_different_hashes(self):
        obj   = Strang("head::tail.a.b.<uuid>")
        obj2  = Strang("head::tail.a.b.<uuid>")
        assert(obj is not obj2)
        assert(str(obj) is not str(obj2))
        assert(hash(obj) != hash(obj2))

class TestStrang_LT:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    ##--| <

    def test_lt(self):
        obj   = Strang("head::tail.a.b.c")
        obj2  = Strang("head::tail.a.b.c.d")
        assert(obj < obj2 )

    def test_lt_mark(self):
        obj   = Strang("head::tail.a.b..c")
        obj2  = Strang("head::tail.a.b..c.d")
        assert(obj < obj2 )

    def test_lt_uuid(self):
        obj   = Strang("head::tail.a.b.c")
        obj2  = Strang("head::tail.a.b.c.<uuid>")
        assert(obj < obj2 )

    def test_lt_fail(self):
        obj   = Strang("head::tail.a.b.c")
        obj2  = Strang("head::tail.a.c.c.d")
        assert(not obj < obj2 )

    def test_lt_fail_on_head(self):
        obj   = Strang("head.blah::tail.a.b.c")
        obj2  = Strang("head::tail.a.b.c.d")
        assert(not obj < obj2 )

    ##--| <=

    def test_le(self):
        obj   = Strang("head::tail.a.b.d")
        obj2  = Strang("head::tail.a.b.d")
        assert(not obj < obj2 )
        assert(obj <= obj2)

    def test_le_on_self(self):
        obj         = Strang("head::tail.a.b.c")
        obj2        = Strang("head::tail.a.b.c")
        assert(obj  == obj2)
        assert(obj  <= obj2 )

    def test_le_on_uuid(self):
        obj  = Strang("head::tail.a.b.c.<uuid>")
        obj2 = Strang(obj[:,:])
        assert(obj.uuid() == obj2.uuid())
        assert(obj == obj2)
        assert(obj <= obj2)

    def test_le_fail_on_uuids(self):
        obj  = Strang("head::tail.a.b.<uuid>")
        obj2 = Strang("head::tail.a.b.<uuid>")
        assert(not obj < obj2 )
        assert(not obj <= obj2)

class TestStrang_Contains:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    ##--| in

    def test_contains(self):
        obj  = Strang("head::tail.a.b.c")
        obj2 = Strang("head::tail.a.b")
        assert(obj2 in obj)

    def test_not_contains(self):
        obj = Strang("head::tail.a.b.c")
        obj2 = Strang("head::tail.a.c.b")
        assert(obj not in obj2)

    def test_contains_word(self):
        obj = Strang("head::tail.a.b.c")
        assert("tail" in obj)

    def test_contains_uuid(self):
        obj = Strang("head::tail.a.b.c.<uuid>")
        match obj.get(1,-1):
            case uuid.UUID() as x:
                assert(x in obj)
            case x:
                assert(False), x

    def test_contains_uuid_fail(self):
        obj = Strang("head::tail.a.b.c.<uuid>")
        assert(not obj.uuid())
        assert(uuid.uuid1() not in obj)

    def test_contains_mark(self):
        obj = Strang("head::tail.a.b.c.$gen$.<uuid>")
        assert(Strang.section('body').marks.gen in obj)

    def test_contains_mark_fail(self):
        obj = Strang("head::tail.a.b.c.$gen$.<uuid>")
        assert(Strang.section('body').marks.head not in obj)

    ##--| match

    def test_match_against_str(self):
        obj  = Strang("head::tail.a.b.c")
        match obj:
            case "head::tail.a.b.c":
                assert(True)
            case _:
                assert(False)

    def test_match_sections(self):
        obj  = Strang("head::tail.a.b.c")
        match obj:
            case Strang(head=x, body=y):
                assert(x == "head")
                assert(y == "tail.a.b.c")
            case x:
                assert(False), x

    def test_match_sections_positionally(self):
        obj  = Strang("head::tail.a.b.c")
        match obj:
            case Strang("head", "tail.a.b.c"):
                assert(True)
            case _:
                assert(False)

    def test_match_sections_literally(self):
        obj  = Strang("head.a.b::tail.a.b.c")
        match obj:
            case Strang(head="head.a.b"):
                assert(True)
            case x:
                assert(False), x

    def test_match_sections_as_str(self):
        obj  = Strang("head.a.b::tail.a.b.c")
        match obj:
            case Strang(head=str() as x):
                assert(x == "head.a.b")
            case x:
                assert(False), x

    def test_match_missing_sections_fail(self):
        obj  = Strang("head::tail.a.b.c")
        match obj:
            case Strang(head=_, tail=_):
                assert(False)
            case _:
                assert(True)

class TestStrang_Modify:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    ##--| Push

    def test_push(self):
        obj = Strang("group::body.a.b.c")
        match obj.push("blah"):
            case Strang(body="body.a.b.c..blah"):
                assert(True)
            case x:
                assert(False), x

    def test_push_none_no_op(self):
        obj = Strang("group::body.a.b.c")
        match obj.push(None):
            case Strang() as x if x == obj:
                assert(True)
            case x:
                assert(False), x

    def test_push_multi(self):
        obj = Strang("group::body.a.b.c")
        match obj.push("d", "e", "f"):
            case Strang(body="body.a.b.c..d.e.f"):
                assert(True)
            case x:
                assert(False), x

    def test_push_multi_with_nones(self):
        obj = Strang("group::body.a.b.c")
        match obj.push("d", "e", None, "f"):
            case Strang(body="body.a.b.c..d.e..f") as x:
                assert(True)
            case x:
                assert(False), x

    def test_push_repeated(self):
        obj = Strang("group::body.a.b.c")
        assert(isinstance((r1:=obj.push("first")), Strang))
        assert(r1 == "group::body.a.b.c..first")
        assert(isinstance((r2:=r1.push("second")), Strang))
        assert(r2 == "group::body.a.b.c..first..second")
        assert(isinstance((r3:=r2.push("third")), Strang))
        assert(r3 == "group::body.a.b.c..first..second..third")
        assert(obj == "group::body.a.b.c")

    def test_push_number(self):
        obj = Strang("group::body.a.b.c")
        for _ in range(10):
            num = randint(0, 100)
            assert(obj.push(num) == f"group::body.a.b.c..{num}")

    ##--| UUIDs

    def test_push_preserves_uuid(self):
        obj = Strang("group::body.a.b.c.<uuid>")
        match obj.push("blah"):
            case Strang() as x:
                assert(x[:] == f"{obj[:]}..blah")
                assert(obj.get(1,4) == x.get(1,4))
            case x:
                assert(False), x

    def test_push_does_not_preserve_uuid_arg(self):
        obj = Strang("group::body.a.b.c[<uuid>]")
        assert(obj.uuid())
        match obj.push("blah"):
            case Strang() as x:
                assert(not x.uuid())
            case x:
                assert(False), x

    def test_push_basic_uuid_str(self):
        obj = Strang("group::body.a.b.c")
        match obj.push("<uuid>"):
            case Strang() as x:
                assert(x[:] == "group::body.a.b.c..<uuid>")
                assert(str(x) == f"group::body.a.b.c..<uuid:{x.get(1,-1)}>")
            case x:
                assert(False), x

    def test_push_explicit_uuid_str(self):
        obj         = Strang("group::body.a.b.c")
        uuid_word   = f"<uuid:{uuid.uuid1()}>"
        ing         = f"group::body.a.b.c..{uuid_word}"
        match obj.push(uuid_word):
            case Strang() as x:
                assert(str(x) == ing)
            case x:
                assert(False), x

    def test_push_uuid_object(self):
        obj       = Strang("group::body.a.b.c")
        new_uuid  = uuid.uuid1()
        match obj.push(new_uuid):
            case Strang(body="body.a.b.c..<uuid>") as x:
                assert(x.get(1,-1) == new_uuid)
            case x:
                assert(False), x

    ##--| Marking

    def test_push_mark(self):
        obj = Strang("group::body")
        mark = obj.section(-1).marks.head
        match obj.push(mark):
            case Strang(body="body..$head$") as x:
                assert(mark in x)
                assert(True)
            case x:
                assert(False), x

    def test_push_mark_idempotent(self):
        obj = Strang("group::body")
        mark = obj.section(-1).marks.head
        match obj.push(mark):
            case Strang(body="body..$head$") as x:
                repeat = x.push(mark)
                assert(mark in x)
                assert(mark in repeat)
                assert(repeat == x)
            case x:
                assert(False), x

    def test_push_mark_idempotent_alt(self):
        obj = Strang("group::body")
        mark = obj.section(-1).marks.head
        match obj.push(mark, mark, mark):
            case Strang(body="body..$head$") as x:
                assert(mark in x)
            case x:
                assert(False), x

    def test_push_mark_repeat(self):
        obj   = Strang("group::body")
        mark  = obj.section(-1).marks.extend
        match obj.push(mark, mark, mark):
            case Strang(body="body..+.+.+") as x:
                assert(True)
            case x:
                assert(False), x

    def test_mark_from_str(self):
        obj   = Strang("group::body")
        match obj.mark("$head$"):
            case Strang() as x:
                assert(x[1,-1] == "$head$")
            case x:
                assert(False), x

    def test_mark_literal(self):
        obj   = Strang("group::body")
        match obj.mark(API.DefaultBodyMarks_e.head):
            case Strang() as x:
                assert(x[1,-1] == "$head$")
            case x:
                assert(False), x

    def test_mark_fail(self):
        obj   = Strang("group::body")
        with pytest.raises(ValueError):
            obj.mark("badval")

    ##--| Pop

    def test_pop_no_marks(self):
        obj = Strang("group::body.a.b.c")
        match obj.pop():
            case Strang(body="body.a.b.c") as x:
                assert(x == obj)
                assert(x is obj)
            case x:
                assert(False), x

    def test_pop_mark(self):
        obj = Strang("group::body.a.b.c..d")
        assert(isinstance((result:=obj.pop()), Strang))
        assert(result == "group::body.a.b.c")
        assert(obj == "group::body.a.b.c..d")

    def test_pop_to_top(self):
        obj = Strang("group::body.a.b.c..d..e")
        assert(isinstance((result:=obj.pop(top=True)), Strang))
        assert(result == "group::body.a.b.c")
        assert(obj == "group::body.a.b.c..d..e")

    def test_pop_to_top_with_markers(self):
        obj = Strang("group::+.body.a.b.c..$head$..<uuid>")
        assert(isinstance((result:=obj.pop(top=True)), Strang))
        assert(result == "group::+.body.a.b.c")

class TestStrang_UUIDs:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_implicit(self):
        obj = Strang("group::body.a.b.c..<uuid>")
        match obj.get(1,-1):
            case uuid.UUID():
                assert(not obj.uuid())
            case x:
                assert(False), x

    def test_implicit_str(self):
        ing = "group::body.a.b.c..<uuid>"
        ang = Strang(ing)
        full_ing = f"group::body.a.b.c..<uuid:{ang.get(1,-1)}>"
        assert(not ang.uuid())
        assert(str(ang) == full_ing)
        assert(ang[:] == ing)
        assert(ang[:,:] == full_ing)

    def test_explicit(self):
        uid_obj = uuid.uuid1()
        ing = f"group::body.a.b.c..<uuid:{uid_obj}>"
        ang = Strang(ing)
        assert(isinstance(ang.get(1,-1), uuid.UUID))
        assert(not ang.uuid())

    def test_explicit_str(self):
        uid_obj = uuid.uuid1()
        ing = f"group::body.a.b.c..<uuid:{uid_obj}>"
        ang = Strang(ing)
        assert(not ang.uuid())
        assert(ang.get(1,-1) == uid_obj)

    def test_uuid_carries_to_new_instances(self):
        uid_obj = uuid.uuid1()
        ing = f"group::body.a.b.c..<uuid:{uid_obj}>"
        ang1 = Strang(ing)
        ang2 = Strang(ang1)
        assert(ang1[1,-1] == ang2[1, -1])

    def test_unique_uuid_carries_to_new_instances(self):
        uuid_obj = uuid.uuid1()
        ing = f"group..<uuid:{uuid_obj}>::body.a.b.c"
        ang1 = Strang(ing)
        ang2 = Strang(ang1)
        assert(ang1[0,-1] == ang2[0, -1])

    def test_error_on_multiple_head_uuids(self):
        ing       = "group::a.b.c[<uuid>,<uuid>]"
        with pytest.raises(StrangError):
            Strang(ing)

    ##--| to unique

    def test_to_uniq(self):
        obj = Strang("group::body.a.b.c")
        match obj.to_uniq():
            case Strang() as x:
                assert(x.uuid())
            case x:
                raise TypeError(type(x))

    def test_to_uniq_with_suffix(self):
        obj = Strang("group::body.a.b.c")
        match obj.to_uniq("blah"):
            case Strang() as x:
                assert(x[:] == "group::body.a.b.c.blah[<uuid>]")
                assert(x.uuid())
            case x:
                assert(False), x

    def test_to_uniq_pop_returns_self(self):
        obj              = Strang("group::body.a.b.c")
        r1               = obj.to_uniq()
        assert(r1 is not obj)
        assert(r1.uuid())
        assert(r1.pop()  == r1)

    def test_to_uniq_idempotent(self):
        obj = Strang("group::body.a.b.c")
        r1  = obj.to_uniq()
        r2  = r1.to_uniq()
        assert(r1.uuid())
        assert(r2.uuid())
        assert(obj != r1)
        assert(obj != r2)
        assert(r1 == r2)
        assert(r1.uuid() == r2.uuid())

    def test_de_uniq(self):
        obj     = Strang("group::body.a.b.c[<uuid>]")
        target  = "group::body.a.b.c"
        assert(obj.uuid())
        match obj.de_uniq():
            case Strang() as x:
                assert(not x.uuid())
                assert(x == target)
                assert(f"{x:a=}" == "")
                assert(True)
            case x:
                assert(False), x

    def test_is_unique(self):
        obj = Strang("head::tail.a.b.c[<uuid>]")
        assert(obj.uuid())

    def test_is_unique_with_uuid(self):
        obj = Strang("head::tail.a.b.c.<uuid>[<uuid>]")
        assert(obj.uuid())
        assert(obj.uuid() != obj.get(1,-1))

    def test_is_not_unique(self):
        obj = Strang("head::tail.a.b.c")
        assert(not obj.uuid())

    def test_is_not_unique_alt(self):
        obj = Strang("head::tail.a.b.c.<uuid>")
        assert(not obj.uuid())

    def test_not_is_unique_alt(self):
        obj = Strang("head::tail.a.b.c.<uuid>")
        assert(not obj.uuid())

    def test_popped_uniq_is_not_uniq(self):
        obj     = Strang("head::tail.a.b.c..<uuid>")
        popped  = obj.pop()
        assert(popped.shape == (1,4))
        assert(not obj.uuid())
        assert(not popped.uuid())
        assert(obj != popped)

class TestStrang_Formatting:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    ##--| repr

    def test_repr_with_brace_val(self):
        obj = Strang("head::tail.{aval}.blah")
        assert(repr(obj) == "<Strang: head::tail.{aval}.blah>")

    def test_repr_uses_no_expansion(self):
        raw               = "head::tail.<uuid>.a.b.c"
        obj               = Strang(raw)
        assert(repr(obj)  == f"<Strang: {obj[:]}>")

    ##--| str

    def test_str_is_full_expansion(self):
        raw             =  "head::tail.<uuid>.a.b.c"
        obj             = Strang(raw)
        body_uuid       = f"<uuid:{obj.get(1,1)}>"
        full_expansion  = f"head::tail.{body_uuid}.a.b.c"
        assert(str(obj) == full_expansion)

    ##--| str slice

    def test_basic_slice_no_expansion(self):
        raw            = "head::tail.<uuid>.a.b.c"
        obj            = Strang(raw)
        assert(obj[:]  == raw)

    def test_basic_slice_but_args_no_expansion(self):
        raw            = "head::tail.<uuid>.a.b.c[<uuid>]"
        obj            = Strang(raw)
        assert(obj[:]  == raw)

    ##--| slice+

    def test_advanced_slice_expands_with_no_args(self):
        raw              = "head::tail.<uuid>.a.b.c"
        obj              = Strang(raw)
        body_uuid        = f"<uuid:{obj.get(1,1)}>"
        expanded         = f"head::tail.{body_uuid}.a.b.c"
        assert(obj[:,:]  == expanded)

    def test_advanced_slice_expands_with_args(self):
        raw              = "head::tail.<uuid>.a.b.c[<uuid>]"
        obj              = Strang(raw)
        body_uuid        = f"<uuid:{obj.get(1,1)}>"
        expanded         = f"head::tail.{body_uuid}.a.b.c"
        assert(obj[:,:]  == expanded)

    def test_format_group(self):
        obj = Strang("group.blah::body.a.b.c")
        assert(f"{obj[0,:]}" == "group.blah")

    def test_format_body(self):
        obj = Strang("group.blah::body.a.b.c")
        assert(f"{obj[1,:]}" == "body.a.b.c")

    def test_format_word(self):
        obj = Strang("group.blah::body.a.b.c")
        assert(f"{obj[1,0]}" == "body")

    ##--| 'u' spec

    def test_format_uuid(self):
        raw                =  "head::tail.a.b.c[<uuid>]"
        obj                = Strang(raw)
        obj_uuid           = f"<uuid:{obj.uuid()}>"
        assert(f"{obj:u}"  == obj_uuid)

    ##--| 'a+' spec

    def test_format_full_expansion(self):
        raw                 = "head::tail.<uuid>.a.b.c"
        obj                 = Strang(raw)
        body_uuid           = f"<uuid:{obj.get(1,1)}>"
        expanded            = f"head::tail.{body_uuid}.a.b.c"
        assert(f"{obj:a+}"  == expanded)

    def test_format_args_a_plus(self):
        raw                =  "head::tail.a.b.c[<uuid>]"
        obj                = Strang(raw)
        uuid_obj           = f"<uuid:{obj.uuid()}>"
        assert(f"{obj:a+}"  == f"head::tail.a.b.c[{uuid_obj}]")

    def test_format_no_args_a_plus(self):
        raw                 =  "head::tail.a.b.c"
        obj                 = Strang(raw)
        assert(f"{obj:a+}"  == "head::tail.a.b.c")

    ##--| 'a' spec

    def test_format_args_a_simple(self):
        raw                 =  "head::tail.a.b.c[<uuid>]"
        obj                 = Strang(raw)
        assert(f"{obj:a}"  == "head::tail.a.b.c[<uuid>]")

    def test_format_no_args_a_simple(self):
        raw                 =  "head::tail.a.b.c"
        obj                 = Strang(raw)
        assert(f"{obj:a}"  == "head::tail.a.b.c")

    ##--| 'a-' spec

    def test_format_args_a_minus(self):
        raw                 =  "head::tail.a.b.c[<uuid>]"
        obj                 = Strang(raw)
        assert(f"{obj:a-}"  == "head::tail.a.b.c")

    def test_format_no_args_a_minus(self):
        raw                 =  "head::tail.a.b.c"
        obj                 = Strang(raw)
        assert(f"{obj:a-}"  == "head::tail.a.b.c")

    ##--| 'a=' spec

    def test_format_args_a_equal(self):
        raw                 =  "head::tail.a.b.c[<uuid>]"
        obj                 = Strang(raw)
        assert(f"{obj:a=}"  == "<uuid>")

    def test_format_no_args_a_equal(self):
        raw                 =  "head::tail.a.b.c"
        obj                 = Strang(raw)
        assert(f"{obj:a=}"  == "")

    ##--| joined

    def test_full_slice_includes_uuid(self):
        raw             =  "head::tail.<uuid>.a.b.c[<uuid>]"
        obj             = Strang(raw)
        body_uuid       = f"<uuid:{obj.get(1,1)}>"
        obj_uuid        = f"<uuid:{obj.uuid()}>"
        basic           =  "head::tail.<uuid>.a.b.c"
        expanded        = f"head::tail.{body_uuid}.a.b.c"
        simple_args     =  "head::tail.<uuid>.a.b.c[<uuid>]"
        full_expansion  = f"head::tail.{body_uuid}.a.b.c[{obj_uuid}]"

        assert(obj.uuid())
        assert(obj[:]   == raw)
        assert(obj[:,:]     == expanded)
        assert(f"{obj:u}"   == obj_uuid)
        assert(f"{obj:a-}"  == basic)
        assert(f"{obj:a}"   == simple_args)
        assert(f"{obj:a+}"  == full_expansion)
        assert(str(obj)     == full_expansion)

class TestStrang_Annotation:
    """ Test custom parameterized subclassing

    """

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_unannotated(self):
        assert(Strang.cls_annotation() == ())
        obj = Strang("group::body")
        assert(obj.cls_annotation() == ())

    def test_simple_annotation(self):
        Strang._clear_registry()
        obj = Strang[int]
        assert(obj is not None)
        assert(obj.__origin__ is Strang)
        assert(obj.__args__ == (int,))

    def test_subclass_annotation(self):
        Strang._clear_registry()
        assert(not bool(Strang._registry))
        assert(isinstance(Strang[int], GenericAlias))

        class IntStrang(Strang[int]):
            __slots__ = ()
            pass

        assert(bool(Strang._registry))
        assert(id(Strang.__annotations__) != id(IntStrang.__annotations__))
        assert(IntStrang.cls_annotation()   == (int,))
        assert(issubclass(IntStrang, Strang))
        assert(issubclass(IntStrang, Strang[int])) # type: ignore[misc]
        assert(IntStrang is Strang[int])

    def test_type_annotated(self):

        Strang._clear_registry()

        class IntStrang(Strang[int]):
            __slots__ = ()
            pass

        assert(issubclass(IntStrang, Strang))
        assert(IntStrang.cls_annotation() == (int,))
        inst = IntStrang("blah::a.b.c")
        assert(inst.cls_annotation() == (int,))
        assert(not hasattr(inst, "__dict__"))

    def test_match_on_strang(self):

        Strang._clear_registry()

        class IntStrang(Strang[int]):
            __slots__ = ()
            pass

        match IntStrang("group.a.b::body.c.d"):
            case Strang(body="body.c.d"):
                assert(True)
            case _:
                assert(False)

    def test_match_on_literal(self):

        Strang._clear_registry()

        class IntStrang(Strang[int]):
            __slots__ = ()
            pass

        match IntStrang("group.a.b::body.c.d"):
            case "group.a.b::body.c.d":
                assert(True)
            case _:
                assert(False)

    def test_match_on_str(self):

        Strang._clear_registry()

        class IntStrang(Strang[int]):
            __slots__ = ()
            pass

        match IntStrang("group.a.b::body.c.d"):
            case str():
                assert(True)
            case _:
                assert(False)

    def test_match_on_subtype(self):
        Strang._clear_registry()

        class IntStrang(Strang[int]):
            __slots__ =()
            pass

        subtype = Strang[int]
        match IntStrang("group.a.b::body.c.d"):
            case x if isinstance(x, subtype):
                assert(True)
            case _:
                assert(False)

    def test_match_on_subtype_fail(self):
        Strang._clear_registry()

        class IntStrang(Strang[int]):
            __slots__ =()
            pass

        match Strang("group.a.b::body.c.d"):
            case IntStrang() as x:
                assert(False), type(x)
            case x:
                assert(True)

    def test_match_on_subtype_fail_b(self):
        Strang._clear_registry()
        cls     = Strang[int]
        notcls  = Strang[float]
        with pytest.raises(TypeError):
            match Strang[int]("group.a.b::body.c.d"):
                case notcls(): # type: ignore[misc]
                    assert(False)
                case cls(): # type: ignore[misc]
                    assert(True)
                case _:
                    assert(False)

class TestStrang_Subclassing:
    """ Check some basic variations of Strang Subclasses,
    like changing the sections

    """

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_three_sections(self) -> None:

        class ThreeSections(Strang):
            """ A strang with 3 sections """
            __slots__ = ()
            _sections : ClassVar = API.Sections_d(
                # name, case, end, types, marks, required
                ("first", ".", "::", str, None, True),
                ("second", "/", ":|:", str, None, True),
                ("third", ".", None, str, None, True),
            )

        match ThreeSections("a.b.c::d/e/f:|:g"):
            case ThreeSections() as val:
                assert(not hasattr(val, "__dict__"))
                assert(val.first == "a.b.c")
                assert(val.second == "d/e/f")
                assert(val.third == "g")
            case x:
                assert(False), x
        assert(issubclass(ThreeSections, Strang))
        assert(isinstance(ThreeSections, API.Strang_p))

    def test_three_sections_errors_on_malformed(self) -> None:

        class ThreeSections(Strang):
            """ A strang with 3 sections """
            __slots__ = ()
            _sections : ClassVar = API.Sections_d(
                # name, case, end, types, marks, required
                ("first", ".", "::", str, None, True),
                ("second", "/", ":|:", str, None, True),
                ("third", ".", None, str, None, True),
            )

        assert(issubclass(ThreeSections, Strang))
        # Check it  errors on malformed
        with pytest.raises(StrangError):
            ThreeSections("a.b.c::d.e.f")

    def test_single_section(self) -> None:

        class OneSection(Strang):
            """ A strang with one section """
            __slots__ = ()
            _sections : ClassVar = API.Sections_d(
                # name, case, end, types, marks, required
                ("first", ".", None, str, None, True),
            )

        assert(issubclass(OneSection, Strang))
        match OneSection("a.b.c::d.e.f"):
            case OneSection() as x:
                assert(x.first == "a.b.c::d.e.f")
                assert(True)
            case x:
                assert(False), x

    def test_end_section(self) -> None:

        class EndSectionStrang(Strang):
            """ A strang with one section """
            __slots__ = ()
            _sections : ClassVar = API.Sections_d(
                # name, case, end, types, marks, required
                ("first", ".", "::", str, None, True),
                ("second", ".", "$", str, None, True),
            )

        assert(issubclass(EndSectionStrang, Strang))
        with pytest.raises(StrangError):
            EndSectionStrang("a.b.c::d.e.f")

        match EndSectionStrang("a.b.c::d.e.f$"):
            case EndSectionStrang() as x:
                assert(x == "a.b.c::d.e.f$")
                assert(True)
            case x:
                assert(False), x

    def test_optional_section(self) -> None:

        class OptSectionStrang(Strang):
            """ A strang with one section """
            __slots__ = ()
            _sections : ClassVar = API.Sections_d(
                # name, case, end, types, marks, required
                ("first", ".", "::",  str, None, True),
                ("second", ".", "::", str, None, False),
                ("third", ".", "$",   str, None, True),
            )

        assert(issubclass(OptSectionStrang, Strang))
        ang1 = OptSectionStrang("a.b.c::d.e.f$")
        ang2 = OptSectionStrang("a.b.c::j.k.l::d.e.f$")

        assert(ang1.first == ang2.first)
        assert(ang1.second == "")
        assert(ang2.second == "j.k.l")
        assert(ang1.third == ang2.third)

    def test_subclass_annotate(self) -> None:

        class StrangSub(Strang):
            __slots__ = ()
            _sections : ClassVar = API.Sections_d(
                # name, case, end, types, marks, required
                ("first", ".", ":|:",  str, None, True),
                ("third", ".", None,   str, None, True),
            )

        ref = StrangSub[int]("group.a.b:|:body.c.d")
        assert(ref.cls_annotation() == ())
        assert(isinstance(ref, Strang))
        assert(isinstance(ref, StrangSub))

    def test_subclass_annotate_independence(self) -> None:

        class StrangSub(Strang):
            __slots__ = ()
            _sections : ClassVar = API.Sections_d(
                # name, case, end, types, marks, required
                ("first", ".", ":|:",  str, None, True),
                ("third", ".", None,   str, None, True),
            )

        ref = StrangSub[int]("group.a.b:|:body.c.d")
        assert(ref.cls_annotation() == ())
        assert(isinstance(ref, Strang))
        assert(isinstance(ref, StrangSub))

        obj : Strang = Strang("group::tail.a.b.c")
        assert(isinstance(obj, Strang))
        assert(not isinstance(obj, StrangSub))

    def test_subclass_matches_protocol(self) -> None:

        class StrangSub(Strang):
            __slots__ = ()
            _sections : ClassVar = API.Sections_d(
                # name, case, end, types, marks, required
                ("first", ".", ":|:",  str, None, True),
                ("third", ".", None,   str, None, True),
            )

        match StrangSub("group:|:tail.a.b.c"):
            case API.Strang_p():
                assert(True)
            case x:
                assert(False), x

    def test_subclass_matches_strang(self) -> None:

        class StrangSub(Strang):
            __slots__ = ()
            _sections : ClassVar = API.Sections_d(
                # name, case, end, types, marks, required
                ("first", ".", ":|:",  str, None, True),
                ("third", ".", None,   str, None, True),
            )

        match StrangSub("group:|:tail.a.b.c"):
            case Strang():
                assert(True)
            case x:
                assert(False), x

class TestStrang_Args:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_slice_with_args(self):
        obj = Strang("head::tail.a.b.c[blah,bloo,blee]")
        assert(isinstance(obj, Strang))
        assert(isinstance(obj, str))
        assert(obj[1,-1] == "c")

    def test_with_no_args(self):
        obj = Strang("head::tail.a.b.c")
        assert(isinstance(obj, Strang))
        assert(isinstance(obj, str))
        assert(obj[1,-1] == "c")

    def test_uuid_arg(self):
        obj = Strang("head::tail.a.b.c[<uuid>]")
        assert(isinstance(obj, Strang))
        assert(isinstance(obj, str))
        assert(obj[1,-1] == "c")
        assert(obj.uuid())

    def test_basic_slice_includes_args(self):
        obj = Strang("head::tail.a.b.c[<uuid>]")
        assert(obj.uuid())
        assert(obj[:] == "head::tail.a.b.c[<uuid>]")

    def test_full_slice_includes_uuid(self):
        raw             =  "head::tail.<uuid>.a.b.c[<uuid>]"
        obj             = Strang(raw)
        body_uuid       = f"<uuid:{obj.get(1,1)}>"
        obj_uuid        = f"<uuid:{obj.uuid()}>"
        basic           =  "head::tail.<uuid>.a.b.c"
        expanded        = f"head::tail.{body_uuid}.a.b.c"
        simple_args     =  "head::tail.<uuid>.a.b.c[<uuid>]"
        full_expansion  = f"head::tail.{body_uuid}.a.b.c[{obj_uuid}]"

        assert(obj.uuid())
        assert(obj[:]   == raw)
        assert(obj[:,:]     == expanded)
        assert(f"{obj:u}"   == obj_uuid)
        assert(f"{obj:a-}"  == basic)
        assert(f"{obj:a}"   == simple_args)
        assert(f"{obj:a+}"  == full_expansion)
        assert(str(obj)     == full_expansion)


    def test_args(self):
        obj = Strang("head.a.b::tail.c.d[blah]")
        match obj.args():
            case ["blah"]:
                assert(True)
            case x:
                assert(False), x

    def test_multi_args(self):
        obj = Strang("head.a.b::tail.c.d[blah, bloo]")
        match obj.args():
            case ["blah", "bloo"]:
                assert(True)
            case x:
                assert(False), x

    def test_empty_args(self):
        obj = Strang("head.a.b::tail.c.d")
        match obj.args():
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_args_to_uniq(self):
        obj = Strang("head.a.b::tail.c.d[blah]")
        uniq = obj.to_uniq()
        match uniq.args():
            case ["<uuid>", "blah"]:
                assert(True)
            case x:
                assert(False), x


    def test_popped_uniq_retains_args(self):
        obj = Strang("head.a.b::tail.c.d[blah]")
        uniq = obj.to_uniq()
        popped = uniq.pop(top=True)
        match popped.args():
            case ["<uuid>", "blah"]:
                assert(True)
            case x:
                assert(False), x



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

##-- end builtin imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import pytest
import os
import tempfile
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator

@pytest.fixture
def wrap_tmp(tmp_path:pl.Path) -> Generator[pl.Path]:
    """ create a new temp directory, and change cwd to it,
      returning to original cwd after the test
      """
    logging.debug("Moving to temp dir")
    orig     = pl.Path().cwd()
    new_base = tmp_path / "test_root"
    new_base.mkdir()
    os.chdir(new_base)
    yield new_base
    logging.debug("Returning to original dir")
    os.chdir(orig)



"""
DKey, a str extension for doing things with str format expansion

"""
from ._interface      import Key_p, DKeyMark_e
from ._util._interface import ExpInst_d
from .errors          import DKeyError
from .dkey            import DKey
# from ._util.formatter import DKeyFormatter
from ._util.decorator import DKeyed, DKeyExpansionDecorator

from .keys import SingleDKey, MultiDKey, NonDKey, IndirectDKey

from .special.import_key     import ImportDKey
from .special.args_keys      import ArgsDKey, KwargsDKey
from .special.str_key        import StrDKey
from .special.path_key       import PathDKey



#!/usr/bin/env python3
"""

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest

# ##-- end 3rd party imports

# ##-- 1st party imports
from .. import param_spec as Specs
from ..arg_parser import CLIParser, ParseMachine
from ..param_spec import ParamSpec

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

# Logging:
logging = logmod.root

# Global Vars:

# Body:
class TestMachine:

    def test_sanity(self):
        assert(True is not False)

    def test_creation(self):
        machine = ParseMachine()
        assert(machine is not None)
        assert(isinstance(machine.model, CLIParser))

    def test_with_custom_model(self):

        class SubParser(CLIParser):
            pass

        machine = ParseMachine(parser=SubParser())
        assert(machine is not None)
        assert(isinstance(machine.model, SubParser))

    def test_setup_ransition(self):
        machine = ParseMachine()
        assert(machine.current_state.id == "Start")
        machine.setup(["a","b","c","d"], None, None, None)
        assert(machine.current_state.id == "Head")

    def test_setup_with_no_more_args(self):
        machine = ParseMachine()
        assert(machine.current_state.id == "Start")
        machine.setup([], None, None, None)
        assert(machine.current_state.id == "ReadyToReport")

    def test_parse_transition(self):
        machine = ParseMachine()
        machine.current_state = machine.Head
        assert(machine.current_state.id == "Head")
        machine.parse()
        assert(machine.current_state.id == "ReadyToReport")

    def test_finish_transition(self):
        machine = ParseMachine()
        machine.current_state = machine.ReadyToReport
        assert(machine.current_state.id == "ReadyToReport")
        machine.finish()
        assert(machine.current_state.id == "End")

    def test_setup_too_many_attempts(self):
        machine = ParseMachine()
        machine.max_attempts = 1
        with pytest.raises(StopIteration):
            machine.setup(["doot", "test"], [], [], [])

    def test_parse_too_many_attempts(self):
        machine = ParseMachine()
        machine.current_state = machine.Head
        machine.model._remaining_args = [1,2,3,4]
        machine.max_attempts = 1
        with pytest.raises(StopIteration):
            machine.parse(["doot", "test"], [], [], [])

    def test_finish_too_many_attempts(self):
        machine = ParseMachine()
        machine.current_state = machine.ReadyToReport
        machine.max_attempts = 1
        with pytest.raises(StopIteration):
            machine.finish()



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import functools as ftz
import inspect
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import sys
import time
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
import jgdv.errors
from jgdv.debugging import TraceBuilder
from jgdv.mixins.annotate import Subclasser
# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
import typing
from typing import cast, assert_type, assert_never
from typing import Generic, NewType
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
from . import _interface as API  # noqa: N812
from ._interface import Signature, Decorable, Decorated, DForm_e, Decorator_p

if typing.TYPE_CHECKING:
    import types
    import enum
    from typing import Final, ClassVar, Any, LiteralString
    from typing import Never, Self, Literal, TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, Either, Func
    from jgdv._abstract.types import Method

##--|
# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# TODO use ideas from pytest.mark
# TODO use strang for mark/data keys
ProtoMeta : Final[type] = type(Protocol)
#--|

class DecoratorMeta(ProtoMeta):

    @overload
    def __call__[T](cls:type[API.Decorator_p], target:type[T], *args:Any, **kwargs:Any) -> type[T]: ...  # noqa: ANN401, N805

    @overload
    def __call__(cls:type[API.Decorator_p], *args:Any, **kwargs:Any) -> type[API.Decorator_p]: ...  # noqa: ANN401, N805

    def __call__(cls, *args:Any, **kwargs:Any):  # noqa: N805
        """ When called with a class as the first arg, builds and calls the decorator on it """
        dec : API.Decorator_p
        dec = cls.__new__(cls)
        assert(isinstance(dec, cls))
        dec.__init__(*args, **kwargs) # type: ignore[misc]
        return dec

class _DecAnnotate_m:
    """ Utils for manipulating annotations related to the decorator
    Annotations for a decorator are stored in a dict entry.
    of the form: '{annotation_prefix}:{data_suffix}'
    """

    def data_key(self:Decorator_p) -> str:
        if not self._data_key:
            self._data_key = f"{self._annotation_prefix}:{self._data_suffix}"

        assert(self._data_key is not None)
        return self._data_key

    def annotate_decorable(self:Decorator_p, target:Decorable) -> list:
        """
        Essentially: target[data_key] += self.{data_key}[:]
        """
        current = target.__annotations__.get(self.data_key(), [])
        match self._build_annotations_h(target, current):
            case []:
                # No Annotations to add
                return []
            case [*xs]:
                logging.info("Applying Annotations to: %s", target)
                target.__annotations__[self.data_key()] = xs
                return xs
            case x:
                msg = "Bad annotation type"
                raise TypeError(msg, x)

    def get_annotations(self:Decorator_p, target:Decorable) -> list[str]:
        """ Get the annotations of the target """
        if not hasattr(target, API.ATTR_TARGET):
            return []
        bottom  = self._unwrap(target)
        data    = bottom.__annotations__.get(self.data_key(), [])
        return data[:]

    def is_annotated(self:Decorator_p, target:Decorable) -> bool:
        logging.info("Testing for annotation data: %s : %s", self.data_key(), target)
        match target:
            case x if not hasattr(x, API.ATTR_TARGET):
                return False
            case type():
                return self.data_key() in target.__annotations__
            case _:
                return self.data_key() in target.__annotations__

class _DecMark_m:
    """ For Marking and checking Decorables.
    Marks are for easily testing if Decorator decorated something already

    """

    def mark_key(self:Decorator_p) -> str:
        if not self._mark_key:
            self._mark_key = f"{self._annotation_prefix}:{self._mark_suffix}"

        assert(self._mark_key is not None)
        return self._mark_key

    def apply_mark(self:Decorator_p, *args:Decorable) -> None:
        """ Mark the UNWRAPPED, original target as already decorated """
        logging.info("Applying Mark %s to : %s", self.mark_key(), args)
        for x in args:
            x.__annotations__[self.mark_key()] = True

    def is_marked(self:Decorator_p, target:Decorable) -> bool:
        logging.info("Testing for mark: %s : %s", self.mark_key(), target)
        match target:
            case x if not hasattr(x, API.ATTR_TARGET):
                return False
            case type() as x:
                return self.mark_key() in x.__annotations__
            case x:
                local_key = self.mark_key() in x.__annotations__
                return local_key or self.is_marked(type(target))

class _DecWrap_m:
    """ Utils for unwrapping and wrapping a  """

    def _unwrap(self:Decorator_p, target:Decorated) -> Decorable:
        """ Get the un-decorated function if there is one """
        match target:
            case type():
                return target
            case x:
                return inspect.unwrap(x)

    def _unwrapped_depth(self:Decorator_p, target:Decorated) -> int:
        """ the code of inspect.unwrap, but used for counting the unwrap depth """
        logging.info("Counting Wrap Depth of: %s", target)
        f               = target
        memo            = {id(f): f}
        depth           = 0
        recursion_limit = sys.getrecursionlimit()
        while not isinstance(f, type) and hasattr(f, API.WRAPPED):
            f = f.__wrapped__ # type: ignore[attr-defined]
            depth += 1
            id_func = id(f)
            if (id_func in memo) or (len(memo) >= recursion_limit):
                msg = f'wrapper loop when unwrapping {target!r}'
                raise ValueError(msg)
            memo[id_func] = f
        else:
            return depth

    def _build_wrapper[**I,O](self:Decorator_p, form:DForm_e, target:Decorable[I,O]) -> Maybe[Decorated[I,O]]:
        """ Create a new decoration using the appropriate hook """
        match form:
            case self.Form.CLASS:
                logging.info("Decorating class: %s", target)
                # Classes are a special case, Maybe modifying instead of wrapping
                assert(isinstance(target, type))
                return cast("Maybe[Decorated[I,O]]", self._wrap_class_h(target))
            case self.Form.METHOD:
                logging.info("Decorating Method: %s", target)
                # TODO if its actually a method type, will need to get the unbound fn
                return self._wrap_method_h(cast("types.MethodType", target))
            case self.Form.FUNC:
                logging.info("Decorating Function: %s", target)
                return self._wrap_fn_h(target)
            case x:
                msg = "Unexpected Decorable type"
                raise ValueError(msg, x)

    def _apply_onto(self:Decorator_p, wrapper:Decorated, target:Decorable) -> Decorated:
        """ Uses functools.update_wrapper,
        Modify cls._wrapper_assignments and cls._wrapper_updates as necessary
        """
        assert(wrapper is not None)
        logging.info("Applying wrapper to decorable: %s -> %s", wrapper, target)
        match target:
            case type():
                return wrapper
            case x:
                return ftz.update_wrapper(wrapper, x,
                                          assigned=self._wrapper_assignments,
                                          updated=self._wrapper_updates)

class _DecInspect_m:

    def _signature(self:Decorator_p, target:Decorable) -> Signature:
        return inspect.signature(target, follow_wrapped=False)

    def _discrim_form(self:Decorator_p, target:Decorable) -> DForm_e:
        """ Determine the type of the thing being decorated"""
        try:
            target = self._unwrap(target)
            if inspect.isclass(target):
                return self.Form.CLASS
            if inspect.ismethod(target):
                return self.Form.METHOD
            if inspect.ismethodwrapper(target):
                return self.Form.METHOD

            # A heuristic Fallback
            match self._signature(target).parameters.get("self", False):
                case False:
                    return self.Form.FUNC
                case _:
                    return self.Form.METHOD
        except TypeError as err:
            raise TypeError(*err.args) from None
        else:
            msg = "Unknown decoration target type"
            raise TypeError(msg, target)

class _DecoratorHooks_m:
    """ The main hooks used to actually specify the decoration """
    _builder : ClassVar[Subclasser] = Subclasser()

    def _wrap_method_h[**In, Out](self:Decorator_p, meth:Method[In,Out]) -> Decorated[In, Out]:
        """ Override this to add a decoration function to method """
        dec_name = self.dec_name()

        def _default_method_wrapper(*args:In.args, **kwargs:In.kwargs) -> Out:
            logging.debug("Calling Wrapped Method: %s of %s", meth.__qualname__, dec_name)
            return meth(*args, **kwargs)

        return cast("Method", _default_method_wrapper)

    def _wrap_fn_h[**In, Out](self:Decorator_p, fn:Func[In, Out]) -> Decorated[In, Out]:
        """ override this to add a decorator to a function """
        dec_name = self.dec_name()

        def _default_fn_wrapper(*args:In.args, **kwargs:In.kwargs) -> Out:
            logging.debug("Calling Wrapped Fn: %s : %s", fn.__qualname__, dec_name)
            return fn(*args, **kwargs)

        return cast("Method", _default_fn_wrapper)

    def _wrap_class_h[**I,O](self, cls:type[O]) -> Maybe[Decorated[I,O]]:
        """ Override this to decorate a class """
        return self._builder.make_subclass("DefaultWrappedClass", cls)

    def _validate_target_h(self:Decorator_p, target:Decorable, form:DForm_e, args:Maybe[list]=None) -> None:
        """ Abstract class for specialization.
        Given the original target, throw an error here if it isn't 'correct' in some way
        """
        pass

    def _validate_sig_h(self:Decorator_p, sig:Signature, form:DForm_e, args:Maybe[list]=None) -> None:
        pass

    def _build_annotations_h(self, target:Decorable, current:list) -> list:  # noqa: ARG002
        """ Given a list of the current annotation list,
        return its replacement
        """
        return []

##--|

class _DecoratorCombined_m(_DecAnnotate_m, _DecWrap_m, _DecMark_m, _DecInspect_m, _DecoratorHooks_m):
    """ Combines the util mixins """
    pass

##--|

class _DecIdempotentLogic_m:
    """ Decorate the passed target in an idempotent way """
    pass

##--|

class Decorator(_DecoratorCombined_m, Decorator_p, metaclass=DecoratorMeta): # type: ignore[misc]
    """
    The abstract Superclass of Decorators
    A subclass implements '_decoration_logic'
    """
    Form                 : ClassVar[type[DForm_e]] = DForm_e
    needs_args           : ClassVar[bool]          = False


    def __init__(self, *args:Any, prefix:Maybe[str]=None, mark:Maybe[str]=None, data:Maybe[str]=None) -> None:  # noqa: ANN401, ARG002
        # Ignores any args
        # TODO use strangs for mark and data key
        self._annotation_prefix   = prefix  or API.ANNOTATIONS_PREFIX
        self._mark_suffix         = mark    or self.__class__.__name__
        self._data_suffix         = data    or API.DATA_SUFFIX
        self._wrapper_assignments = list(ftz.WRAPPER_ASSIGNMENTS)
        self._wrapper_updates     = list(ftz.WRAPPER_UPDATES)
        self._mark_key            = None # type: ignore[assignment]
        self._data_key            = None # type: ignore[assignment]

    def __call__(self, target:Decorable) -> Decorated:
        try:
            decorated = self._decoration_logic(target)
        except Exception as err:  # noqa: BLE001
            # Capture all decoration exceptions,
            # and turn them into JGDVErrors,
            # So the traceback can be manipulated
            raise err.with_traceback(TraceBuilder()[1:]) from None
        else:
            assert(decorated is not None)
            return decorated

    def _decoration_logic[**I, O](self, target:Decorable[I,O]) -> Decorated[I,O]:
        """
        # need to wrap with my wrapper
        annotations = self.get_annotations(target)
        form, sig   = self._discrim_form(target), self._signature(target)
        # Verify the target, may raise exceptions
        self._validate_target_h(target, form, annotations)
        self._validate_sig_h(sig, form, annotations)
        """
        raise NotImplementedError()

    def dec_name(self) -> str:
        return self.__class__.__qualname__

##--|

class MonotonicDec(Decorator):
    """ The Base Monotonic Decorator

    Applying the decorator repeatedly adds successive decoration functions
    Monotonic's don't annotate
    """

    def _decoration_logic[**I, O](self, target:Decorable[I,O]) -> Decorated[I,O]:
        top, bottom = target, self._unwrap(target)
        form, sig = self._discrim_form(bottom), self._signature(bottom)

        self._validate_target_h(bottom, form)
        self._validate_sig_h(sig, form)
        match self._build_wrapper(form, bottom):
            case None:
                return top
            case wrapper if wrapper is top:
                return top
            case wrapper:
                self.apply_mark(wrapper, top, bottom)
                return self._apply_onto(wrapper, top)

class IdempotentDec(Decorator):
    """ The Base Idempotent Decorator

    Already decorated targets are 'marked' with _mark_key as an attr.

    Can annotate targets with metadata without modifying the runtime behaviour,
    or modify the runtime behaviour

    annotations are assigned as fn.__annotations[decorator._data_key] = []
    the mark as fn.__annotations__[decorator._mark_key] = True

    Moving data from wrapped to wrapper is taken care of,
    so no need for ftz.wraps in _wrap_method_h or _wrap_fn_h

    """

    def _decoration_logic(self, target:Decorable) -> Decorated:
        top, bottom = target, self._unwrap(target)
        match self.is_marked(bottom):
            case True if top is not bottom:
                # Already wrapped, nothing to do
                return top
            case True:
                msg = "A Marked Decorable doesn't have a wrapper"
                raise ValueError(msg, target)
            case False:
                form = self._discrim_form(bottom)

        match self._build_wrapper(form, bottom):
            case None:
                # Decorable was modified
                return top
            case type() as x:
                self.apply_mark(x, top, bottom)
                return x
            case wrapper if wrapper is top:
                self.apply_mark(wrapper, top, bottom)
                return top
            case wrapper:
                self.apply_mark(wrapper, top, bottom)
                return self._apply_onto(wrapper, top)

class MetaDec(Decorator):
    """
    Adds metadata without modifying runtime behaviour of target,
    Or validates a class

    ie: annotates without wrapping
    """

    def __init__(self, value:str|list[str], **kwargs) -> None:  # noqa: ANN003
        kwargs.setdefault("mark", "_meta_marked")
        kwargs.setdefault("data", "_meta_vals")
        super().__init__(**kwargs)
        match value:
            case list():
                self._data = value
            case _:
                self._data = [value]

    def _decoration_logic(self, target:Decorable) -> Decorated:
        top, bottom = target, self._unwrap(target)
        form, sig   = self._discrim_form(target), self._signature(bottom)
        annotations = self.annotate_decorable(bottom)
        # Verify the target, may raise exceptions
        self._validate_target_h(bottom, form, annotations)
        self._validate_sig_h(sig, form, annotations)
        return top

    def _build_annotations_h(self, target:Decorable, current:list) -> list:  # noqa: ARG002
        return [*current, *self._data]

class DataDec(IdempotentDec):
    """
    An extended IdempotentDec, which uses a data annotation
    on the original Decorable,
    to run the single wrapping function
    """

    def __init__(self, keys:str|list[str], **kwargs) -> None:  # noqa: ANN003
        kwargs.setdefault("mark", "_d_marked")
        kwargs.setdefault("data", "_d_vals")
        super().__init__(**kwargs)
        match keys:
            case list():
                self._data = keys
            case _:
                self._data = [keys]

    def _decoration_logic(self, target:Decorable) -> Decorated:
        top, bottom = target, self._unwrap(target)
        match self.annotate_decorable(bottom):
            case []:
                # No annotations added
                return top
            case list() as annots if top is not bottom and self.is_marked(bottom):
                # Theres a wrapper, and its mine
                # Verify the target, may raise exceptions
                form, sig = self._discrim_form(target), self._signature(bottom)
                self._validate_target_h(bottom, form, annots)
                self._validate_sig_h(sig, form, annots)
                return top
            case list() as annots:
                form, sig = self._discrim_form(target), self._signature(bottom)
                self._validate_target_h(bottom, form, annots)
                self._validate_sig_h(sig, form, annots)
                # Now handle the wrapping
                return super()._decoration_logic(top)
            case x:
                raise TypeError(type(x))

    def _build_annotations_h(self, target:Decorable, current:list) -> list:  # noqa: ARG002
        return [*self._data, *current]



#!/usr/bin/env python3
"""
Utility classes for attribute based access to loaded toml data ::

    # Simplifies:
    data['blah']['awe']['awg']
    # to:
    data.blah.awe.awg

    # Also allows guarded access:

    result = data.on_fail('fallback').somewhere.along.this.doesnt.exist()
    result == "fallback" or data.somewhere.along.this.doesnt.exist


The underlying Python access model (simplified)::

    object.__getattribute(self, name):
        try:
            return self.__dict__[name]
        except AttributeError:
            return self.__getattr__(name)

So by looking up values in ChainGuard.__table and handling missing values,
we can skip dict style key access

"""
from typing import Final, TypeAlias
import datetime
from collections import ChainMap

# ##-- Generated Exports
__all__ = (
    "ChainGuard",
    "load"
    "GuardedAccessError",
)
# ##-- end Generated Exports


from .errors import GuardedAccessError
from ._interface import TomlTypes
from .chainguard import ChainGuard

load        = ChainGuard.load     # type: ignore
load_dir    = ChainGuard.load_dir # type: ignore
read        = ChainGuard.read     # type: ignore



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN202, B011, PLR0133
# Imports:
from __future__ import annotations

import itertools as itz
import logging as logmod
import warnings
import pathlib as pl
import pytest
from jgdv.cli import ParseError
from jgdv.cli.param_spec import ParamSpec
from jgdv.cli.param_spec.defaults import HelpParam, VerboseParam, SeparatorParam

##--| Values
logging = logmod.root
good_names = ("test", "blah", "bloo")
bad_names  = ("-test", "blah=bloo")

##--| Tests

class TestHelpParam:

    def test_sanity(self):
        assert(True is not False)

    def test_basic(self):
        match HelpParam():
            case HelpParam() as val:
                assert(val.default is False)
                assert(val.implicit is True)
            case x:
                assert(False), x

    def test_consume(self):
        obj = HelpParam()
        in_data = ["--help"]
        match obj.consume(in_data):
            case {"help": True}, 1:
                assert(True)
            case x:
                assert(False), x

    def test_consume_short(self):
        obj = HelpParam()
        in_data = ["--h"]
        match obj.consume(in_data):
            case {"help": True}, 1:
                assert(True)
            case x:
                assert(False), x

class TestVerboseParam:

    def test_sanity(self):
        assert(True is not False)

    def test_basic(self):
        match VerboseParam():
            case VerboseParam() as val:
                assert(val.implicit is True)
            case x:
                assert(False), x

class TestSeparatorParam:

    def test_sanity(self):
        assert(True is not False)

    def test_basic(self):
        match SeparatorParam():
            case SeparatorParam() as val:
                assert(val.implicit is True)
            case x:
                assert(False), x


    def test_consume(self):
        in_args = ["--"]
        obj = SeparatorParam()
        assert(obj.prefix == "-")
        assert(obj.name == "-")
        match obj.consume(in_args):
            case {"-": True}, 1:
                assert(True)
            case x:
                assert(False), x
